import unittest

from qmctorch.sampler import GeneralizedMetropolis
from .test_sampler_base import TestSamplerBase


class TestGeneralizeMetropolis(TestSamplerBase):

    def test_gmh(self):
        """Test generalized MH."""
        sampler = GeneralizedMetropolis(
            nwalkers=10, nstep=20, step_size=0.2,
            nelec=self.wf.nelec, ndim=self.wf.ndim,
            init=self.mol.domain('normal'))

        pos = sampler(self.wf.pdf)


if __name__ == "__main__":
    unittest.main()



import torch

from .atomic_orbitals import AtomicOrbitals
from .backflow.orbital_dependent_backflow_transformation import OrbitalDependentBackFlowTransformation


class AtomicOrbitalsOrbitalDependentBackFlow(AtomicOrbitals):

    def __init__(self, mol, backflow_kernel, backflow_kernel_kwargs={}, cuda=False):
        """Computes the value of atomic orbitals

        Args:
            mol (Molecule): Molecule object
            cuda (bool, optional): Turn GPU ON/OFF Defaults to False.
        """

        super().__init__(mol, cuda)
        dtype = torch.get_default_dtype()
        self.backflow_trans = OrbitalDependentBackFlowTransformation(mol,
                                                                     backflow_kernel=backflow_kernel,
                                                                     backflow_kernel_kwargs=backflow_kernel_kwargs,
                                                                     cuda=cuda)

    def forward(self, pos, derivative=[0], sum_grad=True, sum_hess=True, one_elec=False):
        r"""Computes the values of the atomic orbitals.

        .. math::
            \phi_i(r_j) = \sum_n c_n \\text{Rad}^{i}_n(r_j) \\text{Y}^{i}_n(r_j)

        where Rad is the radial part and Y the spherical harmonics part.
        It is also possible to compute the first and second derivatives

        .. math::
            \\nabla \phi_i(r_j) = \\frac{d}{dx_j} \phi_i(r_j) + \\frac{d}{dy_j} \phi_i(r_j) + \\frac{d}{dz_j} \phi_i(r_j) \n
            \\text{grad} \phi_i(r_j) = (\\frac{d}{dx_j} \phi_i(r_j), \\frac{d}{dy_j} \phi_i(r_j), \\frac{d}{dz_j} \phi_i(r_j)) \n
            \Delta \phi_i(r_j) = \\frac{d^2}{dx^2_j} \phi_i(r_j) + \\frac{d^2}{dy^2_j} \phi_i(r_j) + \\frac{d^2}{dz^2_j} \phi_i(r_j)

        Args:
            pos (torch.tensor): Positions of the electrons
                                  Size : Nbatch, Nelec x Ndim
            derivative (int, optional): order of the derivative (0,1,2,).
                                        Defaults to 0.
            sum_grad (bool, optional): Return the sum_grad (i.e. the sum of
                                       the derivatives) or the individual
                                       terms. Defaults to True.
                                       False only for derivative=1

            one_elec (bool, optional): if only one electron is in input

        Returns:
            torch.tensor: Value of the AO (or their derivatives) \n
                          size : Nbatch, Nelec, Norb (sum_grad = True) \n
                          size : Nbatch, Nelec, Norb, Ndim (sum_grad = False)

        Examples::
            >>> mol = Molecule('h2.xyz')
            >>> ao = AtomicOrbitals(mol)
            >>> pos = torch.rand(100,6)
            >>> aovals = ao(pos)
            >>> daovals = ao(pos,derivative=1)
        """

        if not isinstance(derivative, list):
            derivative = [derivative]

        if not sum_grad:
            assert(1 in derivative)

        if not sum_hess:
            assert(2 in derivative)

        if one_elec:
            nelec_save = self.nelec
            self.nelec = 1

        if derivative == [0]:
            ao = self._compute_ao_values(pos)

        elif derivative == [1]:
            ao = self._compute_first_derivative_ao_values(
                pos, sum_grad)

        elif derivative == [2]:
            ao = self._compute_second_derivative_ao_values(
                pos, sum_hess)

        elif derivative == [3]:
            ao = self._compute_mixed_second_derivative_ao_values(pos)

        elif derivative == [0, 1, 2]:
            ao = self._compute_all_backflow_ao_values(pos)

        else:
            raise ValueError(
                'derivative must be 0, 1, 2 or [0, 1, 2], got ', derivative)

        if one_elec:
            self.nelec = nelec_save

        return ao

    def _compute_first_derivative_ao_values(self, pos, sum_grad):
        """Compute the value of the derivative of the ao from the xyx and r tensor

        Args:
            pos (torch.tensor): position of each elec size Nbatch, NelexNdim
            sum_grad (boolean): return the sum_grad (True) or gradient (False)

        Returns:
            torch.tensor: derivative of atomic orbital values
                          size (Nbatch, Nelec, Norb) if sum_grad
                          size (Nbatch, Nelec, Norb, Ndim) if sum_grad=False
        """
        grad = self._compute_gradient_backflow_ao_values(pos)

        if sum_grad:
            _, nbatch, nelec, norb = grad.shape
            grad = grad.reshape(nelec, 3, nbatch, nelec, norb).sum(1)

        return grad

    def _compute_gradient_backflow_ao_values(self, pos, grad_ao=None):
        """Compute the jacobian of the backflow ao fromn xyz tensor

        Args:
            pos ([type]): [description]

        Returns:
            torch.tensor derivative of atomic orbital values
                          size ([Nelec*Ndim], Nbatch, Nelec, Norb)
        """

        # get the grad vals : size Nbatch x Nelec x Norb x Ndim
        # the derivative are : d \phi_i(q_j) / d \tilde{x}_j
        # i.e. wrt to the back flow coordinates
        if grad_ao is None:
            grad_ao = self._compute_gradient_ao_values(pos)

        # permute the grad to Nbatch x Ndim x 1 x Nelec x 1 x Norb
        grad_ao = grad_ao.permute(0, 3, 1, 2)
        grad_ao = grad_ao.unsqueeze(2).unsqueeze(4)

        # compute the derivative of the bf positions wrt to the original pos
        # Nbatch x Ndim x Ndim x Nelec x Nelec x Norb
        dbf = self.backflow_trans(
            pos, derivative=1).permute(0, 2, 3, 4, 5, 1)

        # compute backflow : Nbatch x Ndim x Nelec x Nelec x Norb
        grad_ao = (grad_ao * dbf).sum(1)

        # permute to have Nelec x Ndim x Nbatch x Nelec x Norb
        # Note the idex of the electron (2,3) are inverted
        # compare to the single backflow trans case
        grad_ao = grad_ao.permute(3, 1, 0, 2, 4)

        # collapse the first two dim [Nelec*Ndim] x Nbatch x Nelec x Norb
        grad_ao = grad_ao.reshape(-1, *(grad_ao.shape[2:]))

        return grad_ao

    def _compute_second_derivative_ao_values(self, pos, sum_hess):
        """Compute the value of the 2nd derivative of the ao from the xyx and r tensor

        Args:
            pos (torch.tensor): position of each elec size Nbatch, NelexNdim
            sum_hess (boolean): return the sum_hess (True) or diag hess (False)

        Returns:
            torch.tensor: derivative of atomic orbital values
                          size (Nbatch, Nelec, Norb) if sum_grad
                          size (Nbatch, Nelec, Norb, Ndim) if sum_grad=False
        """
        hess = self._compute_diag_hessian_backflow_ao_values(pos)

        if sum_hess:
            _, nbatch, nelec, norb = hess.shape
            hess = hess.reshape(nelec, 3, nbatch, nelec, norb).sum(1)

        return hess

    def _compute_diag_hessian_backflow_ao_values(self, pos, hess_ao=None, mixed_ao=None, grad_ao=None):
        """Compute the laplacian of the backflow ao fromn xyz tensor

        Args:
            pos ([type]): [description]

        Returns:
            torch.tensor derivative of atomic orbital values
                          size (Nbatch, Nelec, Nelec, Norb)
        """
        # compute the lap size Nbatch x Ndim x Ndim x Nelec x Norb
        if hess_ao is None:
            hess_ao = self._compute_diag_hessian_ao_values(pos)

        if mixed_ao is None:
            mixed_ao = self._compute_mixed_second_derivative_ao_values(
                pos)

        if grad_ao is None:
            grad_ao = self._compute_gradient_ao_values(pos)

        # permute the grad to Nbatch x Ndim x 1 x Nelec x 1 x Norb
        grad_ao = grad_ao.permute(0, 3, 1, 2)
        grad_ao = grad_ao.unsqueeze(2).unsqueeze(4)

        # permute the hess to Nbatch x Ndim x 1 x Nelec x 1 x Norb
        hess_ao = hess_ao.permute(0, 3, 1, 2)
        hess_ao = hess_ao.unsqueeze(2).unsqueeze(4)

        # permute the hess to Nbatch x Ndim x 1 x Nelec x 1 x Norb
        mixed_ao = mixed_ao.permute(0, 3, 1, 2)
        mixed_ao = mixed_ao.unsqueeze(2).unsqueeze(4)

        # compute the derivative of the bf positions wrt to the original pos
        # Nbatch x Ndim x Ndim x Nelec x Nelec x Norb
        dbf = self.backflow_trans(
            pos, derivative=1).permute(0, 2, 3, 4, 5, 1)

        # compute the derivative of the bf positions wrt to the original pos
        # Nbatch x Ndim x Ndim x Nelec x Nelec x Norb
        d2bf = self.backflow_trans(
            pos, derivative=2).permute(0, 2, 3, 4, 5, 1)

        # compute the back flow second der
        hess_ao = (hess_ao * (dbf*dbf)).sum(1)

        # compute the backflow grad
        hess_ao += (grad_ao * d2bf).sum(1)

        # compute the contribution of the mixed derivative
        hess_ao += 2*(mixed_ao *
                      dbf[:, [[0, 1], [0, 2], [1, 2]], ...].prod(2)).sum(1)

        # permute to have Nelec x Ndim x Nbatch x Nelec x Norb
        hess_ao = hess_ao.permute(3, 1, 0, 2, 4)

        # collapse the first two dim [Nelec*Ndim] x Nbatch x Nelec x Norb
        hess_ao = hess_ao.reshape(-1, *(hess_ao.shape[2:]))

        return hess_ao

    def _compute_all_backflow_ao_values(self, pos):
        """Compute the ao, gradient, laplacian of the ao from the xyx and r tensor

        Args:
            pos (torch.tensor): position of each elec size Nbatch, Nelec x Ndim

        Returns:
            tuple(): (ao, grad and lapalcian) of atomic orbital values
                     ao size (Nbatch, Nelec, Norb)
                     dao size (Nbatch, Nelec, Norb, Ndim)
                     d2ao size (Nbatch, Nelec, Norb)

        """

        xyz, r = self._process_position(pos)

        R, dR, d2R = self.radial(r, self.bas_n, self.bas_exp,
                                 xyz=xyz, derivative=[0, 1, 2],
                                 sum_grad=False, sum_hess=False)

        Y, dY, d2Y = self.harmonics(xyz,
                                    derivative=[0, 1, 2],
                                    sum_grad=False, sum_hess=False)

        # vals of the bf ao
        ao = self._ao_kernel(R, Y)

        # grad kernel of the bf ao
        grad_ao = self._gradient_kernel(R, dR, Y, dY)

        # diag hess kernel of the bf ao
        hess_ao = self._diag_hessian_kernel(
            R, dR, d2R, Y, dY, d2Y)

        # compute the bf ao
        hess_ao = self._compute_diag_hessian_backflow_ao_values(
            pos, hess_ao=hess_ao, grad_ao=grad_ao)

        # compute the bf grad
        grad_ao = self._compute_gradient_backflow_ao_values(
            pos, grad_ao=grad_ao)

        return (ao, grad_ao, hess_ao)

    def _process_position(self, pos):
        """Computes the positions/distance bewteen elec/orb

        Args:
            pos (torch.tensor): positions of the walkers Nbat, NelecxNdim

        Returns:
            torch.tensor, torch.tensor: positions of the elec wrt the bas
                                        (Nbatch, Nelec, Norb, Ndim)
                                        distance between elec and bas
                                        (Nbatch, Nelec, Norb)
        """
        # get the elec-atom vectrors/distances
        xyz, r = self._elec_ao_dist(pos)

        if self.contract:
            # repeat/interleave to get vector and distance between
            # electrons and orbitals
            xyz = xyz.repeat_interleave(self.nctr_per_ao, dim=2)
            r = r.repeat_interleave(self.nctr_per_ao, dim=2)

        return xyz, r

    def _elec_ao_dist(self, pos):
        """Computes the positions/distance bewteen elec/atoms

        Args:
            pos (torch.tensor): positions of the walkers Nbat, NelecxNdim

        Returns:
            torch.tensor, torch.tensor: positions of the elec wrt the bas
                                        (Nbatch, Nelec, Natom, Ndim)
                                        distance between elec and bas
                                        (Nbatch, Nelec, Natom)
        """

        # compute the back flow positions
        # Nbatch x Nao x Nelec*Ndim
        bf_pos = self.backflow_trans(pos)
        nbatch, nao, _ = bf_pos.shape

        # reshape
        bf_pos = bf_pos.view(nbatch, nao, self.nelec, self.ndim)

        # permute to nbatch x nelec x nao x ndim
        bf_pos = bf_pos.permute(0, 2, 1, 3)

        # interleave the atomic positions
        # nao x ndim
        atom_coords = self.atom_coords.repeat_interleave(
            self.nao_per_atom, dim=0)

        # compute the vectors between electrons and atoms
        # nbatch x nelec x nao x ndim
        xyz = (bf_pos-atom_coords)

        # distance between electrons and atoms
        # nbatch x nelec x nao
        r = torch.sqrt((xyz*xyz).sum(3))

        return xyz, r



#!/usr/bin/env python

import os

from setuptools import (find_packages, setup)

here = os.path.abspath(os.path.dirname(__file__))

# To update the package version number, edit QMCTorch/__version__.py
version = {}
with open(os.path.join(here, 'qmctorch', '__version__.py')) as f:
    exec(f.read(), version)

with open('README.md') as readme_file:
    readme = readme_file.read()

setup(
    name='qmctorch',
    version=version['__version__'],
    description="Pytorch Implementation of Quantum Monte Carlo",
    long_description=readme + '\n\n',
    long_description_content_type='text/markdown',
    author=["Nicolas Renaud", "Felipe Zapata"],
    author_email='n.renaud@esciencecenter.nl',
    url='https://github.com/NLESC-JCER/QMCTorch',
    packages=find_packages(),
    package_dir={'qmctorch': 'qmctorch'},
    include_package_data=True,
    license="Apache Software License 2.0",
    zip_safe=False,
    keywords='qmctorch',
    scripts=['bin/qmctorch'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering :: Chemistry'
    ],
    test_suite='tests',
    install_requires=['matplotlib', 'numpy', 'argparse',
                      'scipy', 'tqdm', 'torch', 'plams',
                      'pyscf', 'mendeleev', 'twiggy', 'mpi4py'],

    extras_require={
        'hpc': ['horovod==0.27.0'],
        'doc': ['recommonmark', 'sphinx', 'sphinx_rtd_theme', 'nbsphinx'],
        'test': ['pytest', 'pytest-runner',
                 'coverage', 'coveralls', 'pycodestyle'],
    }
)



from qmctorch.scf import Molecule
from qmctorch.wavefunction import SlaterManyBodyJastrow
from qmctorch.utils import set_torch_double_precision
from qmctorch.wavefunction.jastrows.elec_nuclei.kernels import PadeJastrowKernel as PadeJastrowKernelElecNuc
from qmctorch.wavefunction.jastrows.elec_elec.kernels import PadeJastrowKernel as PadeJastrowKernelElecElec
from qmctorch.wavefunction.jastrows.elec_elec_nuclei.kernels import BoysHandyJastrowKernel, FullyConnectedJastrowKernel
from torch.autograd import grad, gradcheck, Variable

import numpy as np
import torch
import unittest

torch.set_default_tensor_type(torch.DoubleTensor)


def hess(out, pos):
    # compute the jacobian
    z = Variable(torch.ones(out.shape))
    jacob = grad(out, pos,
                 grad_outputs=z,
                 only_inputs=True,
                 create_graph=True)[0]

    # compute the diagonal element of the Hessian
    z = Variable(torch.ones(jacob.shape[0]))
    hess = torch.zeros(jacob.shape)

    for idim in range(jacob.shape[1]):

        tmp = grad(jacob[:, idim], pos,
                   grad_outputs=z,
                   only_inputs=True,
                   create_graph=True)[0]

        hess[:, idim] = tmp[:, idim]

    return hess


class TestSlaterCombinedJastrow(unittest.TestCase):

    def setUp(self):

        torch.manual_seed(101)
        np.random.seed(101)

        set_torch_double_precision()

        # molecule
        mol = Molecule(
            atom='Li 0 0 0; H 0 0 3.14',
            unit='bohr',
            calculator='pyscf',
            basis='sto-3g',
            redo_scf=True)

        self.wf = SlaterManyBodyJastrow(mol,
                                        kinetic='auto',
                                        include_all_mo=False,
                                        configs='single_double(2,2)',
                                        jastrow_kernel={
                                            'ee': PadeJastrowKernelElecElec,
                                            'en': PadeJastrowKernelElecNuc,
                                            'een': BoysHandyJastrowKernel},
                                        jastrow_kernel_kwargs={
                                            'ee': {'w': 1.},
                                            'en': {'w': 1.},
                                            'een': {}})

        self.random_fc_weight = torch.rand(self.wf.fc.weight.shape)
        self.wf.fc.weight.data = self.random_fc_weight
        self.nbatch = 11
        self.pos = torch.Tensor(np.random.rand(
            self.nbatch,  self.wf.nelec*3))
        self.pos.requires_grad = True

    def test_forward(self):
        _ = self.wf(self.pos)

    def test_antisymmetry(self):
        """Test that the wf values are antisymmetric
        wrt exchange of 2 electrons of same spin."""
        wfvals_ref = self.wf(self.pos)

        if self.wf.nelec < 4:
            print(
                'Warning : antisymmetry cannot be tested with \
                    only %d electrons' % self.wf.nelec)
            return

        # test spin up
        pos_xup = self.pos.clone()
        perm_up = list(range(self.wf.nelec))
        perm_up[0] = 1
        perm_up[1] = 0
        pos_xup = pos_xup.reshape(self.nbatch, self.wf.nelec, 3)
        pos_xup = pos_xup[:, perm_up, :].reshape(
            self.nbatch, self.wf.nelec*3)

        wfvals_xup = self.wf(pos_xup)
        assert(torch.allclose(wfvals_ref, -1 * wfvals_xup))

        # test spin down
        pos_xdn = self.pos.clone()
        perm_dn = list(range(self.wf.nelec))
        perm_dn[self.wf.mol.nup-1] = self.wf.mol.nup
        perm_dn[self.wf.mol.nup] = self.wf.mol.nup-1
        pos_xdn = pos_xdn.reshape(self.nbatch, self.wf.nelec, 3)
        pos_xdn = pos_xdn[:, perm_up, :].reshape(
            self.nbatch, self.wf.nelec*3)

        wfvals_xdn = self.wf(pos_xdn)
        assert(torch.allclose(wfvals_ref, -1.*wfvals_xdn))

    def test_grad_mo(self):
        """Gradients of the MOs."""

        mo = self.wf.pos2mo(self.pos)
        dmo = self.wf.pos2mo(self.pos, derivative=1)

        dmo_grad = grad(
            mo,
            self.pos,
            grad_outputs=torch.ones_like(mo))[0]

        gradcheck(self.wf.pos2mo, self.pos)

        assert(torch.allclose(dmo.sum(), dmo_grad.sum()))
        assert(torch.allclose(dmo.sum(-1),
                              dmo_grad.view(self.nbatch, self.wf.nelec, 3).sum(-1)))

    def test_hess_mo(self):
        """Hessian of the MOs."""
        val = self.wf.pos2mo(self.pos)

        d2val_grad = hess(val, self.pos)
        d2val = self.wf.pos2mo(self.pos, derivative=2)

        assert(torch.allclose(d2val.sum(), d2val_grad.sum()))

        assert(torch.allclose(d2val.sum(-1).sum(-1),
                              d2val_grad.view(self.nbatch, self.wf.nelec, 3).sum(-1).sum(-1)))

        assert(torch.allclose(d2val.sum(-1),
                              d2val_grad.view(self.nbatch, self.wf.nelec, 3).sum(-1)))

    def test_local_energy(self):

        self.wf.kinetic_energy = self.wf.kinetic_energy_autograd
        eloc_auto = self.wf.local_energy(self.pos)

        self.wf.kinetic_energy = self.wf.kinetic_energy_jacobi
        eloc_jac = self.wf.local_energy(self.pos)

        assert torch.allclose(
            eloc_auto.data, eloc_jac.data, rtol=1E-4, atol=1E-4)

    def test_kinetic_energy(self):

        eauto = self.wf.kinetic_energy_autograd(self.pos)
        ejac = self.wf.kinetic_energy_jacobi(self.pos)

        assert torch.allclose(
            eauto.data, ejac.data, rtol=1E-4, atol=1E-4)

    def test_gradients_wf(self):

        grads = self.wf.gradients_jacobi(
            self.pos, sum_grad=False).squeeze()
        grad_auto = self.wf.gradients_autograd(self.pos)

        assert torch.allclose(grads.sum(), grad_auto.sum())

        grads = grads.reshape(self.nbatch, self.wf.nelec, 3)
        grad_auto = grad_auto.reshape(self.nbatch, self.wf.nelec, 3)
        assert(torch.allclose(grads, grad_auto))

    def test_gradients_pdf(self):

        grads_pdf = self.wf.gradients_jacobi(self.pos, pdf=True)
        grads_auto = self.wf.gradients_autograd(self.pos, pdf=True)

        assert torch.allclose(grads_pdf.sum(), grads_auto.sum())


if __name__ == "__main__":
    unittest.main()
    # t = TestSlaterCombinedJastrow()
    # t.setUp()
    # t.test_antisymmetry()
    # t.test_kinetic_energy()



from qmctorch.scf import Molecule
from qmctorch.wavefunction import SlaterOrbitalDependentJastrow
from qmctorch.wavefunction.jastrows.elec_elec.kernels import FullyConnectedJastrowKernel
from qmctorch.utils import set_torch_double_precision, btrace

from torch.autograd import grad, Variable

import numpy as np
import torch
import unittest


def hess(out, pos):
    # compute the jacobian
    z = Variable(torch.ones(out.shape))
    jacob = grad(out, pos,
                 grad_outputs=z,
                 only_inputs=True,
                 create_graph=True)[0]

    # compute the diagonal element of the Hessian
    z = Variable(torch.ones(jacob.shape[0]))
    hess = torch.zeros(jacob.shape)

    for idim in range(jacob.shape[1]):

        tmp = grad(jacob[:, idim], pos,
                   grad_outputs=z,
                   only_inputs=True,
                   create_graph=True)[0]

        hess[:, idim] = tmp[:, idim]

    return hess


class TestSlaterOrbitalDependentJastrow(unittest.TestCase):

    def setUp(self):

        torch.manual_seed(101)
        np.random.seed(101)

        set_torch_double_precision()

        # molecule
        mol = Molecule(
            atom='Li 0 0 0; H 0 0 3.015',
            unit='bohr',
            calculator='pyscf',
            basis='sto-3g',
            redo_scf=True)

        self.wf = SlaterOrbitalDependentJastrow(
            mol,
            kinetic='auto',
            jastrow_kernel=FullyConnectedJastrowKernel,
            configs='single_double(2,4)',
            include_all_mo=True)

        self.random_fc_weight = torch.rand(self.wf.fc.weight.shape)
        self.wf.fc.weight.data = self.random_fc_weight

        self.nbatch = 3
        self.pos = torch.as_tensor(np.random.rand(
            self.nbatch, self.wf.nelec*3))

        self.pos.requires_grad = True

    def test_forward(self):
        """Value of the wave function."""
        wfvals = self.wf(self.pos)
        # ref = torch.as_tensor([[-1.0935e-02], [6.4874e-02], [1.7879e-04],
        #                        [1.5797e-02], [7.4684e-02], [-4.4445e-02],
        #                        [-4.8149e-04], [-3.0355e-03], [-2.0027e-02],
        #                        [5.1957e-05]])
        # assert torch.allclose(wfvals.data, ref, rtol=1E-4, atol=1E-4)

    def test_antisymmetry(self):
        """Test that the wf values are antisymmetric
        wrt exchange of 2 electrons of same spin."""
        wfvals_ref = self.wf(self.pos)

        if self.wf.nelec < 4:
            print(
                'Warning : antisymmetry cannot be tested with \
                    only %d electrons' % self.wf.nelec)
            return

        # test spin up
        pos_xup = self.pos.clone()
        perm_up = list(range(self.wf.nelec))
        perm_up[0] = 1
        perm_up[1] = 0
        pos_xup = pos_xup.reshape(self.nbatch, self.wf.nelec, 3)
        pos_xup = pos_xup[:, perm_up, :].reshape(
            self.nbatch, self.wf.nelec*3)

        wfvals_xup = self.wf(pos_xup)
        assert(torch.allclose(wfvals_ref, -1.*wfvals_xup))

    def test_jacobian_mo(self):
        """Jacobian of the uncorrelated MOs."""
        mo = self.wf.pos2mo(self.pos)
        dmo = self.wf.pos2mo(self.pos, derivative=1)
        dmo_grad = grad(
            mo, self.pos, grad_outputs=torch.ones_like(mo))[0]

        assert(torch.allclose(dmo.sum(-1),
                              dmo_grad.view(self.nbatch, self.wf.nelec, 3).sum(-1)))

    def test_grad_mo(self):
        """Gradients of the uncorrelated MOs."""
        mo = self.wf.pos2mo(self.pos)
        dmo = self.wf.pos2mo(self.pos, derivative=1, sum_grad=False)
        dmo_grad = grad(
            mo, self.pos, grad_outputs=torch.ones_like(mo))[0]

        assert(torch.allclose(dmo.sum(-2),
                              dmo_grad.view(self.nbatch, self.wf.nelec, 3)))

    def test_hess_mo(self):
        """Hessian of the uncorrelated MOs."""
        mo = self.wf.pos2mo(self.pos)
        d2mo = self.wf.pos2mo(self.pos, derivative=2)
        d2mo_grad = hess(mo, self.pos)

        assert(torch.allclose(d2mo.sum(-1),
                              d2mo_grad.view(self.nbatch, self.wf.nelec, 3).sum(-1)))

    def test_jacobian_jast(self):
        """Jacobian of the jastrow values."""
        jast = self.wf.ordered_jastrow(self.pos)
        djast = self.wf.ordered_jastrow(self.pos, derivative=1)
        djast_grad = grad(jast, self.pos,
                          grad_outputs=torch.ones_like(jast))[0]

        assert(torch.allclose(djast_grad.view(self.nbatch, self.wf.nelec, 3).sum(-1),
                              djast.sum(-1)))

    def test_grad_jast(self):
        """Gradients of the jastrow values."""
        jast = self.wf.ordered_jastrow(self.pos)
        djast = self.wf.ordered_jastrow(
            self.pos, derivative=1, sum_grad=False)
        djast_grad = grad(jast, self.pos,
                          grad_outputs=torch.ones_like(jast))[0]

        assert(torch.allclose(djast_grad.view(self.nbatch, self.wf.nelec, 3),
                              djast.sum(-2)))

    def test_hess_jast(self):
        """Hessian of the jastrows."""
        jast = self.wf.ordered_jastrow(self.pos)
        d2jast = self.wf.ordered_jastrow(self.pos, derivative=2)

        d2jast_grad = hess(jast, self.pos)

        assert(torch.allclose(d2jast.sum(-1),
                              d2jast_grad.view(self.nbatch, self.wf.nelec, 3).sum(-1)))

    def test_grad_cmo(self):
        """Gradients of the correlated MOs."""
        cmo = self.wf.pos2cmo(self.pos)
        dcmo = self.wf.get_gradient_operator(self.pos)

        dcmo = dcmo.permute(1, 2, 3, 0)
        shape = (self.nbatch, self.wf.nelec,
                 self.wf.nmo_opt, self.wf.nelec, 3)
        dcmo = dcmo.reshape(*shape)
        dcmo = dcmo.sum(2).sum(1)

        dcmo_grad = grad(cmo, self.pos,
                         grad_outputs=torch.ones_like(cmo))[0]
        dcmo_grad = dcmo_grad.reshape(self.nbatch, self.wf.nelec, 3)

        assert(torch.allclose(dcmo, dcmo_grad))

    def test_hess_cmo(self):
        """Hessian of the correlated MOs."""
        val = self.wf.pos2cmo(self.pos)
        d2val_grad = hess(val, self.pos)

        d2val = self.wf.get_hessian_operator(self.pos)
        d2val = d2val.permute(1, 2, 0, 3).sum(1)

        assert(torch.allclose(d2val.sum(-1),
                              d2val_grad.view(self.nbatch, self.wf.nelec, 3).sum(-1)))

    def test_jacobian_wf(self):
        """Jacobian of det(CMO).
            \nabla det(CMOup) / det(CMOup) +  \nabla det(CMOup) / det(CMOup) """
        grad_jacobi = self.wf.gradients_jacobi(self.pos)
        grad_auto = self.wf.gradients_autograd(self.pos)
        assert(torch.allclose(grad_jacobi, grad_auto.sum(-1)))

    def test_grad_wf(self):
        """Compute the gradients of the wf  wrt to xyz coord of each elec."""
        grad_jacobi = self.wf.gradients_jacobi(
            self.pos, sum_grad=False).squeeze()
        grad_auto = self.wf.gradients_autograd(self.pos)
        assert torch.allclose(grad_jacobi, grad_auto)

    def test_kinetic_energy(self):
        """Kinetic energty."""
        eauto = self.wf.kinetic_energy_autograd(self.pos)
        ejac = self.wf.kinetic_energy_jacobi(self.pos)

        assert torch.allclose(
            eauto.data, ejac.data, rtol=1E-4, atol=1E-4)

    def test_local_energy(self):
        """local energy."""
        self.wf.kinetic_energy = self.wf.kinetic_energy_autograd
        eloc_auto = self.wf.local_energy(self.pos)

        self.wf.kinetic_energy = self.wf.kinetic_energy_autograd
        eloc_jac = self.wf.local_energy(self.pos)

        assert torch.allclose(
            eloc_auto.data, eloc_jac.data, rtol=1E-4, atol=1E-4)


if __name__ == "__main__":
    unittest.main()




import unittest

import numpy as np
import torch
import torch.optim as optim


from qmctorch.sampler import Metropolis
from qmctorch.solver import Solver
from qmctorch.utils import plot_energy


from qmctorch.scf import Molecule
from qmctorch.wavefunction import SlaterOrbitalDependentJastrow
from qmctorch.utils import set_torch_double_precision
from qmctorch.wavefunction.jastrows.elec_elec.kernels import FullyConnectedJastrowKernel
__PLOT__ = False


class TestH2Correlated(unittest.TestCase):

    def setUp(self):

        torch.manual_seed(0)
        np.random.seed(0)
        set_torch_double_precision()

        # optimal parameters
        self.opt_r = 0.69  # the two h are at +0.69 and -0.69
        self.opt_sigma = 1.24

        # molecule
        self.mol = Molecule(
            atom='H 0 0 -0.69; H 0 0 0.69',
            unit='bohr',
            calculator='pyscf',
            basis='sto-3g')

        # wave function
        self.wf = SlaterOrbitalDependentJastrow(self.mol,
                                                kinetic='auto',
                                                configs='cas(2,2)',
                                                jastrow_kernel=FullyConnectedJastrowKernel,
                                                include_all_mo=True)

        # sampler
        self.sampler = Metropolis(
            nwalkers=1000,
            nstep=2000,
            step_size=0.5,
            ndim=self.wf.ndim,
            nelec=self.wf.nelec,
            init=self.mol.domain('normal'),
            move={
                'type': 'all-elec',
                'proba': 'normal'})

        # optimizer
        self.opt = optim.Adam(self.wf.parameters(), lr=0.01)

        # solver
        self.solver = Solver(wf=self.wf, sampler=self.sampler,
                                          optimizer=self.opt)

        # ground state energy
        self.ground_state_energy = -1.16

        # ground state pos
        self.ground_state_pos = 0.69

    def test_0_wavefunction(self):

        # artificial pos
        self.nbatch = 10
        self.pos = torch.as_tensor(np.random.rand(
            self.nbatch, self.wf.nelec*3))
        self.pos.requires_grad = True

        eauto = self.wf.kinetic_energy_autograd(self.pos)
        ejac = self.wf.kinetic_energy_jacobi(self.pos)
        print(torch.stack([eauto, ejac], axis=1).squeeze())
        assert torch.allclose(
            eauto.data, ejac.data, rtol=1E-4, atol=1E-4)

    def test1_single_point(self):

        self.solver.wf.ao.atom_coords[0, 2] = -self.ground_state_pos
        self.solver.wf.ao.atom_coords[1, 2] = self.ground_state_pos
        self.solver.sampler = self.sampler

        # sample and compute observables
        obs = self.solver.single_point()
        e, v = obs.energy, obs.variance

    def test3_wf_opt(self):
        self.solver.sampler = self.sampler
        self.solver.configure(
            track=['local_energy', 'parameters'], loss='energy', grad='auto')
        obs = self.solver.run(5)
        if __PLOT__:
            plot_energy(obs.local_energy, e0=-
                        1.1645, show_variance=True)

    def test4_geo_opt(self):

        self.solver.wf.ao.atom_coords[0,
                                      2].data = torch.as_tensor(-0.37)
        self.solver.wf.ao.atom_coords[1,
                                      2].data = torch.as_tensor(0.37)

        self.solver.configure(track=['local_energy'],
                              loss='energy', grad='auto')
        self.solver.geo_opt(5, nepoch_wf_init=10, nepoch_wf_update=5,
                            hdf5_group='geo_opt_correlated')

        # load the best model
        self.solver.wf.load(self.solver.hdf5file,
                            'geo_opt_correlated')
        self.solver.wf.eval()

        # sample and compute variables
        obs = self.solver.single_point()
        e, v = obs.energy, obs.variance

        e = e.data.numpy()
        v = v.data.numpy()

        # it might be too much to assert with the ground state energy
        assert(e > 2 * self.ground_state_energy and e < 0.)
        assert(v > 0 and v < 2.)


if __name__ == "__main__":
    # unittest.main()
    t = TestH2Correlated()
    t.setUp()
    # t.test_0_wavefunction()
    # t.test1_single_point()
    # t.test2_single_point_hmc()
    t.test3_wf_opt()
    # t.test5_sampling_traj()



import torch
from torch import nn

from .....utils import register_extra_attributes
from .jastrow_kernel_electron_electron_base import JastrowKernelElectronElectronBase


class PadeJastrowPolynomialKernel(JastrowKernelElectronElectronBase):

    def __init__(self, nup, ndown, cuda,
                 order=2,
                 weight_a=None,
                 weight_b=None):
        """Computes a polynomial Pade-Jastrow factor

        .. math::

            B_{ij} =  \\frac{P_{ij}}{Q_{ij}}

        with :
        .. math::

            P_{ij} = a_1 r_{i,j} + a_2 r_{ij}^2 + ....

        and :

        .. math
            Q_{ij} = 1 + b_1 r_{i,j} + b_2 r_{ij}^2 + ...

        Args:
            nup (int): number of spin up electons
            ndow (int): number of spin down electons
            order (int): degree of the polynomial
            weight_a (torch.tensor, optional): Value of the weight
            weight_b (torch.tensor, optional): Value of the weight
            cuda (bool, optional): Turns GPU ON/OFF. Defaults to False.
        """

        super().__init__(nup, ndown, cuda)
        self.porder = order

        self.set_variational_weights(weight_a, weight_b)

        self.static_weight = self.get_static_weight()
        self.requires_autograd = False

    def get_static_weight(self):
        """Get the matrix of static weights

        Returns:
            torch.tensor: static weight (0.5 (0.25) for parallel(anti) spins
        """

        bup = torch.cat((0.25 * torch.ones(self.nup, self.nup), 0.5 *
                         torch.ones(self.nup, self.ndown)), dim=1)

        bdown = torch.cat((0.5 * torch.ones(self.ndown, self.nup), 0.25 *
                           torch.ones(self.ndown, self.ndown)), dim=1)

        static_weight = torch.cat((bup, bdown), dim=0).to(self.device)

        mask_tri_up = torch.triu(torch.ones_like(
            static_weight), diagonal=1).type(torch.BoolTensor).to(self.device)
        static_weight = static_weight.masked_select(mask_tri_up)

        return static_weight

    def set_variational_weights(self, weight_a, weight_b):
        """Define the initial values of the variational weights.

        Args:
            weight_a (torch.tensor or None): Value of the weight
            weight_b (torch.tensor or None): Value of the weight

        """

        # that can cause a nan if too low ...
        w0 = 1E-5

        if weight_a is not None:
            assert weight_a.shape[0] == self.porder
            self.weight_a = nn.Parameter(weight_a)
        else:
            self.weight_a = nn.Parameter(w0 * torch.ones(self.porder))

        if weight_b is not None:
            assert weight_b.shape[0] == self.porder
            self.weight_b = nn.Parameter(weight_b)
        else:
            self.weight_b = nn.Parameter(w0 * torch.ones(self.porder))
            self.weight_b.data[0] = 1.

        register_extra_attributes(self, ['weight_a'])
        register_extra_attributes(self, ['weight_b'])

    def forward(self, r):
        """ Get the jastrow kernel.

        .. math::

            B_{ij} = \\frac{P_{ij}}{Q_{ij}}


        Args:
            r (torch.tensor): matrix of the e-e distances
                              Nbatch x Nelec x Nelec

        Returns:
            torch.tensor: matrix of the jastrow kernels
                          Nbatch x Nelec x Nelec
        """

        num, denom = self._compute_polynoms(r)
        return num / denom

    def compute_derivative(self, r, dr):
        """Get the elements of the derivative of the jastrow kernels
        wrt to the first electrons

        The derivative is given by:

        .. math::

            \\text{out}_{k,i,j} = \\frac{P'Q - PQ'}{Q^2}

        with:

        .. math::
            P_{ij} = a_1 r_{i,j} + a_2 r_{ij}^2 + ....
            Q_{ij} = 1 + b_1 r_{i,j} + b_2 r_{ij}^2 +

        and :

        .. math::

            P'_{ij} = a_1 dr + a_2 2 r dr + a_r 3 dr r^2 + ....
            Q'_{ij} = b_1 dr + b_2 2 r dr + b_r 3 dr r^2 + ....

        Due to the properties of the derivative we have
        .. math::

            \\frac{d B_{ij}}{d k_i} =  \\frac{d B_{ij}}{d k_j}  = -\\frac{d B_{ji}{d k_i}


        Args:
            r (torch.tensor): matrix of the e-e distances
                              Nbatch x Nelec x Nelec
            dr (torch.tensor): matrix of the derivative of the e-e distances
                              Nbatch x Ndim x Nelec x Nelec

        Returns:
            torch.tensor: matrix fof the derivative of the jastrow elements
                          Nbatch x Ndim x Nelec x Nelec
        """

        num, denom = self._compute_polynoms(r)
        num = num.unsqueeze(1)
        denom = denom.unsqueeze(1)

        der_num, der_denom = self._compute_polynom_derivatives(r, dr)

        return (der_num * denom - num * der_denom) / (denom * denom)

    def compute_second_derivative(self, r, dr, d2r):
        """Get the elements of the pure 2nd derivative of the jastrow kernels
        wrt to the first electron

        Due to the properties of the derivative we have
        .. math::

            \\frac{d B_{ij}}{d k_i} =  \\frac{d B_{ij}}{d k_j}  = \\frac{d B_{ji}{d k_i}

        Args:
            r (torch.tensor): matrix of the e-e distances
                              Nbatch x Nelec x Nelec
            dr (torch.tensor): matrix of the derivative of the e-e distances
                              Nbatch x Ndim x Nelec x Nelec
            d2r (torch.tensor): matrix of the 2nd derivative of
                                the e-e distances
                              Nbatch x Ndim x Nelec x Nelec

        Returns:
            torch.tensor: matrix fof the pure 2nd derivative of
                          the jastrow elements
                          Nbatch x Ndim x Nelec x Nelec
        """

        num, denom = self._compute_polynoms(r)
        num = num.unsqueeze(1)
        denom = denom.unsqueeze(1)

        der_num, der_denom = self._compute_polynom_derivatives(r, dr)

        d2_num, d2_denom = self._compute_polynom_second_derivative(
            r, dr, d2r)

        out = d2_num / denom - (2 * der_num * der_denom + num * d2_denom) / (
            denom * denom) + 2 * num * der_denom * der_denom / (denom * denom *
                                                                denom)

        return out

    def _compute_polynoms(self, r):
        """Compute the num and denom polynomials.

        Args:
            r (torch.tensor): matrix of the e-e distances
                              Nbatch x Nelec x Nelec

        Returns:
            torch.tensor, torch.tensor : p and q polynoms
                                         size Nbatch x Nelec x Nelec
        """

        num = self.static_weight * r
        denom = (1.0 + self.weight_b[0] * r)
        riord = r.clone()

        for iord in range(1, self.porder):
            riord = riord * r
            num += self.weight_a[iord] * riord
            denom += self.weight_b[iord] * riord

        return num, denom

    def _compute_polynom_derivatives(self, r, dr):
        """Computes the derivatives of the polynomials.

        Args:
            r (torch.tensor): matrix of the e-e distances
                              Nbatch x Nelec x Nelec
            dr (torch.tensor): matrix of the derivative of the e-e distances
                              Nbatch x Ndim x Nelec x Nelec

        Returns:
            torch.tensor, torch.tensor : p and q polynoms derivatives
                                         size Nbatch x Ndim x Nelec x Nelec

        """

        der_num = self.static_weight * dr
        der_denom = self.weight_b[0] * dr

        r_ = r.unsqueeze(1)
        riord = r.unsqueeze(1)

        for iord in range(1, self.porder):

            fact = (iord + 1) * dr * riord
            der_num += self.weight_a[iord] * fact
            der_denom += self.weight_b[iord] * fact
            riord = riord * r_

        return der_num, der_denom

    def _compute_polynom_second_derivative(self, r, dr, d2r):
        """Computes the second derivative of the polynoms.

        Args:
            r (torch.tensor): matrix of the e-e distances
                              Nbatch x Nelec x Nelec
            dr (torch.tensor): matrix of the derivative of the e-e distances
                              Nbatch x Ndim x Nelec x Nelec
            d2r (torch.tensor): matrix of the 2nd derivative of
                                the e-e distances
                              Nbatch x Ndim x Nelec x Nelec
        Returns:
            torch.tensor, torch.tensor : p and q polynoms derivatives
                                         size Nbatch x Ndim x Nelec x Nelec

        """

        d2_num = self.static_weight * d2r
        d2_denom = self.weight_b[0] * d2r

        dr2 = dr * dr

        r_ = r.unsqueeze(1)
        rnm1 = r.unsqueeze(1)
        rnm2 = 1.

        for iord in range(1, self.porder):

            n = iord + 1
            fact = n * (d2r * rnm1 + iord * dr2 * rnm2)
            d2_num += self.weight_a[iord] * fact
            d2_denom += self.weight_b[iord] * fact

            rnm2 = rnm1
            rnm1 = rnm1 * r_

        return d2_num, d2_denom



import torch
from torch import nn
from torch.autograd import grad, Variable
from torch.utils.data import Dataset


def set_torch_double_precision():
    """Set the default precision to double for all torch tensors."""
    torch.set_default_dtype = torch.float64
    torch.set_default_tensor_type(torch.DoubleTensor)


def set_torch_single_precision():
    """Set the default precision to single for all torch tensors."""
    torch.set_default_dtype = torch.float32
    torch.set_default_tensor_type(torch.FloatTensor)


def fast_power(x, k, mask0=None, mask2=None):
    """Computes x**k when k have elements 0, 1, 2

    Args:
        x (torch.tensor): input
        k (torch.tensor): exponents
        mask0 (torch.tensor): precomputed mask of the elements of that are 0 (Defaults to None and computed here)
        mask2 (torch.tensor): precomputed mask of the elements of that are 2 (Defaults to None and computed here)

    Returns:
        torch.tensor: values of x**k
    """
    kmax = 3
    if k.max() < kmax:

        out = x.clone()

        if mask0 is None:
            mask0 = k == 0

        out.masked_fill_(mask0, 1)

        if k.max() > 1:
            if mask2 is None:
                mask2 = k == 2
            out[..., mask2] *= out[..., mask2]

    else:
        out = x**k

    return out


def gradients(out, inp):
    """Return the gradients of out wrt inp

    Args:
        out ([type]): [description]
        inp ([type]): [description]
    """
    return grad(out, inp, grad_outputs=torch.ones_like(out))


def diagonal_hessian(out, inp, return_grads=False):
    """return the diagonal hessian of out wrt to inp

    Args:
        out ([type]): [description]
        inp ([type]): [description]

    Returns:
        [type]: [description]
    """
    # compute the jacobian
    z = Variable(torch.ones(out.shape))
    jacob = grad(out, inp,
                 grad_outputs=z,
                 only_inputs=True,
                 create_graph=True)[0]

    if return_grads:
        grads = jacob.detach()

    # compute the diagonal element of the Hessian
    z = Variable(torch.ones(jacob.shape[0]))
    hess = torch.zeros(jacob.shape)

    for idim in range(jacob.shape[1]):

        tmp = grad(jacob[:, idim], inp,
                   grad_outputs=z,
                   only_inputs=True,
                   create_graph=True)[0]

        hess[:, idim] = tmp[:, idim]

    if return_grads:
        return hess, grads

    else:
        return hess


class DataSet(Dataset):

    def __init__(self, data):
        """Creates a torch data set

        Arguments:
            data {torch.tensor} -- data
        """
        self.data = data

    def __len__(self):
        """get the number of data points

        Returns:
            int -- number of data points
        """
        return self.data.shape[0]

    def __getitem__(self, index):
        """returns a given data point

        Arguments:
            index {int} -- index of the point

        Returns:
            torch.tensor -- data of that point
        """
        return self.data[index, :]


class Loss(nn.Module):

    def __init__(
            self,
            wf,
            method='energy',
            clip=False):
        """Defines the loss to use during the optimization

        Arguments:
            wf {WaveFunction} -- wave function object used

        Keyword Arguments:
            method {str} -- method to use  (default: {'energy'})
                            (energy, variance, weighted-energy,
                            weighted-variance)
            clip {bool} -- clip the values that are +/- % sigma away from
                           the mean (default: {False})
        """

        super(Loss, self).__init__()

        self.wf = wf
        self.method = method
        self.clip = clip

        # by default we use weights
        # that are needed if we do
        # not resample at every time step
        self.use_weight = True

        # number of +/- std for clipping
        # Excludes values + /- Nstd x std the mean of the eloc
        self.clip_num_std = 5

        # select loss function
        self.loss_fn = {'energy': torch.mean,
                        'variance': torch.var}[method]

        # init values of the weights
        self.weight = {'psi': None, 'psi0': None}

    def forward(self, pos, no_grad=False, deactivate_weight=False):
        """Computes the loss

        Arguments:
            pos {torch.tensor} -- positions of the walkers in that batch

        Keyword Arguments:
            no_grad {bool} -- computes the gradient of the loss
                              (default: {False})

        Returns:
            torch.tensor, torch.tensor -- value of the loss, local energies
        """

        # check if grads are requested
        with self.get_grad_mode(no_grad):

            # compute local eneergies
            local_energies = self.wf.local_energy(pos)

            # mask the energies if necessary
            mask = self.get_clipping_mask(local_energies)

            # sampling_weight
            weight = self.get_sampling_weights(pos, deactivate_weight)

            # compute the loss
            loss = self.loss_fn((weight * local_energies)[mask])

        return loss, local_energies

    @staticmethod
    def get_grad_mode(no_grad):
        """Returns enable_grad or no_grad

        Arguments:
            no_grad {bool} -- [description]
        """

        return torch.no_grad() if no_grad else torch.enable_grad()

    def get_clipping_mask(self, local_energies):
        """computes the clipping mask

        Arguments:
            local_energies {torch.tensor} -- values of the local energies
        """
        if self.clip:
            median = torch.median(local_energies)
            std = torch.std(local_energies)
            emax = median + self.clip_num_std * std
            emin = median - self.clip_num_std * std
            mask = (
                local_energies < emax) & (
                local_energies > emin)
        else:
            mask = torch.ones_like(
                local_energies).type(torch.bool)

        return mask

    def get_sampling_weights(self, pos, deactivate_weight):
        """Get the weight needed when resampling is not
            done at every step
        """

        local_use_weight = self.use_weight * \
            (not deactivate_weight)

        if local_use_weight:

            # computes the weights
            self.weight['psi'] = self.wf(pos)

            # if we just resampled store psi and all w=1
            if self.weight['psi0'] is None:
                self.weight['psi0'] = self.weight['psi'].detach(
                ).clone()
                w = torch.ones_like(self.weight['psi'])

            # otherwise compute ration of psi
            else:
                w = (self.weight['psi'] / self.weight['psi0'])**2
                w /= w.sum()  # should we multiply by the number of elements ?

            return w

        else:
            return 1.


class OrthoReg(nn.Module):
    '''add a penalty to make matrice orthgonal.'''

    def __init__(self, alpha=0.1):
        """Add a penalty loss to keep the MO orthogonalized

        Keyword Arguments:
            alpha {float} -- strength of the penaly (default: {0.1})
        """
        super(OrthoReg, self).__init__()
        self.alpha = alpha

    def forward(self, W):
        """Return the loss : |W x W^T - I|."""
        return self.alpha * \
            torch.norm(W.mm(W.transpose(0, 1)) -
                       torch.eye(W.shape[0]))



from time import time

import numpy as np
import torch
from scipy.interpolate import LinearNDInterpolator, RegularGridInterpolator


class InterpolateMolecularOrbitals:

    def __init__(self, wf):
        """Interpolation of the AO using a log grid centered on each atom."""
        self.wf = wf

    def __call__(self, pos, method='irreg', orb='occupied', **kwargs):

        if method == 'irreg':
            n = kwargs['n'] if 'n' in kwargs else 6
            return self.interpolate_mo_irreg_grid(pos, n=n, orb=orb)

        elif method == 'reg':
            rstr, bstr = 'resolution', 'border_length'
            res = kwargs[rstr] if rstr in kwargs else 0.1
            blength = kwargs[bstr] if bstr in kwargs else 2.
            return self.interpolate_mo_reg_grid(pos, res, blength, orb)

    def get_mo_max_index(self, orb):
        """Get the index of the highest MO to inlcude in the interpoaltion

        Args:
            orb (str): occupied or all

        Raises:
            ValueError: if orb not valid
        """

        if orb == 'occupied':
            self.mo_max_index = torch.stack(
                self.wf.configs).max().item()+1
        elif orb == 'all':
            self.mo_max_index = self.wf.mol.basis.nmo+1
        else:
            raise ValueError(
                'orb must occupied or all')

    def interpolate_mo_irreg_grid(self, pos, n, orb):
        """Interpolate the mo occupied in the configs.

        Args:
            pos (torch.tensor): sampling points (Nbatch, 3*Nelec)
            n (int, optional): Interpolation order. Defaults to 6.

        Returns:
            torch.tensor: mo values Nbatch, Nelec, Nmo
        """
        self.get_mo_max_index(orb)

        if not hasattr(self, 'interp_mo_func'):
            grid_pts = get_log_grid(self.wf.mol.atom_coords, n=n)

            def func(x):
                x = torch.as_tensor(x).type(torch.get_default_dtype())
                ao = self.wf.ao(x, one_elec=True)
                mo = self.wf.mo(self.wf.mo_scf(ao)).squeeze(1)
                return mo[:, :self.mo_max_index].detach()

            self.interp_mo_func = interpolator_irreg_grid(
                func, grid_pts)

        nbatch = pos.shape[0]
        mos = torch.zeros(nbatch, self.wf.mol.nelec,
                          self.wf.mol.basis.nmo)
        mos[:, :, :self.mo_max_index] = interpolate_irreg_grid(
            self.interp_mo_func, pos)
        return mos

    def interpolate_mo_reg_grid(self, pos,  res, blength, orb):
        """Interpolate the mo occupied in the configs.

        Args:
            pos (torch.tensor): sampling points (Nbatch, 3*Nelec)

        Returns:
            torch.tensor: mo values Nbatch, Nelec, Nmo
        """

        self.get_mo_max_index(orb)

        if not hasattr(self, 'interp_mo_func'):
            x, y, z = get_reg_grid(
                self.wf.mol.atom_coords, resolution=res, border_length=blength)

            def func(x):
                x = torch.as_tensor(x).type(torch.get_default_dtype())
                ao = self.wf.ao(x, one_elec=True)
                mo = self.wf.mo(self.wf.mo_scf(ao)).squeeze(1)
                return mo[:, :self.mo_max_index]

            self.interp_mo_func = interpolator_reg_grid(
                func, x, y, z)

        nbatch = pos.shape[0]
        mos = torch.zeros(nbatch, self.wf.mol.nelec,
                          self.wf.mol.basis.nmo)
        mos[:, :, :self.mo_max_index] = interpolate_reg_grid(
            self.interp_mo_func, pos)
        return mos


class InterpolateAtomicOrbitals:

    def __init__(self, wf):
        """Interpolation of the AO using a log grid centered on each atom."""
        self.wf = wf

    def __call__(self, pos, n=6, length=2):
        """Interpolate the AO.

        Args:
            pos (torch.tensor): positions of the walkers
            n (int, optional): number of points on each log axis. Defaults to 6.
            length (int, optional): half length of the grid. Defaults to 2.

        Returns:
            torch.tensor: Interpolated values
        """

        if not hasattr(self, 'interp_func'):
            t0 = time()
            self.get_interpolator()
            print('___', time()-t0)

        t0 = time()
        bas_coords = self.wf.ao.atom_coords.repeat_interleave(
            self.wf.ao.nao_per_atom, dim=0)  # <- we need the number of AO per atom not the number of BAS per atom
        print('___bas ', time()-t0)

        t0 = time()
        xyz = (pos.view(-1, self.wf.ao.nelec, 1, self.wf.ao.ndim) -
               bas_coords[None, ...]).detach().numpy()
        print('___ xyz', time()-t0)

        t0 = time()
        data = np.array([self.interp_func[iorb](xyz[:, :, iorb, :])
                         for iorb in range(self.wf.ao.norb)])
        print('___ data', time()-t0)

        return torch.as_tensor(data.transpose(1, 2, 0))

    def get_interpolator(self, n=6, length=2):
        """evaluate the interpolation function.

        Args:
            n (int, optional): number of points on each log axis. Defaults to 6.
            length (int, optional): half length of the grid. Defaults to 2.
        """

        xpts = logspace(n, length)
        nxpts = len(xpts)

        grid = np.stack(np.meshgrid(
            xpts, xpts, xpts, indexing='ij')).T.reshape(-1, 3)[:, [2, 1, 0]]

        def func(x):
            x = torch.as_tensor(x).type(torch.get_default_dtype())
            nbatch = x.shape[0]
            xyz = x.view(-1, 1, 1, 3).expand(-1,
                                             1, self.wf.ao.nbas, 3)
            r = torch.sqrt((xyz**2).sum(3))
            R = self.wf.ao.radial(r, self.wf.ao.bas_n,
                                  self.wf.ao.bas_exp)
            Y = self.wf.ao.harmonics(xyz)
            bas = R * Y
            bas = self.wf.ao.norm_cst * self.wf.ao.bas_coeffs * bas
            ao = torch.zeros(nbatch, self.wf.ao.nelec,
                             self.wf.ao.norb, device=self.wf.ao.device)
            ao.index_add_(2, self.wf.ao.index_ctr, bas)
            return ao

        data = func(grid).detach().numpy()
        data = data.reshape(nxpts, nxpts, nxpts, -1)

        self.interp_func = [RegularGridInterpolator((xpts, xpts, xpts),
                                                    data[..., i],
                                                    method='linear',
                                                    bounds_error=False,
                                                    fill_value=0.) for i in range(self.wf.ao.norb)]


def get_boundaries(atomic_positions, border_length=2.):
    """Computes the boundaries of the structure

    Args:
        atomic_positions (torch.Tensor, np.ndarray, list): atomic positions
        border_length (float, optional): length of the border. Defaults to 2.

    Raises:
        ValueError: if type of positions not recognized

    Returns:
        (np.ndarray, np.ndarray, mp.ndarray): min, max values in the 3 cartesian directions
    """
    if isinstance(atomic_positions, torch.Tensor):
        pmin = atomic_positions.min(0)[0].detach().cpu().numpy()
        pmax = atomic_positions.max(0)[0].detach().cpu().numpy()

    elif isinstance(atomic_positions, np.ndarray):
        pmin, pmax = atomic_positions.min(0), atomic_positions.max(0)

    elif isinstance(atomic_positions, list):
        _tmp = np.array(atomic_positions)
        pmin, pmax = _tmp.min(0), _tmp.max(0)

    else:
        raise ValueError(
            'atomic_positions must be either a torch.tensor, np.ndarray, or list')

    pmin -= border_length
    pmax += border_length

    return pmin, pmax


def get_reg_grid(atomic_positions, resolution=0.1, border_length=2.):
    """Computes a regular grid points from the atomic positions

    Args:
        atomic_positions (torch.Tensor, np.ndarray, list): atomic positions
        resolution (float, optional): ditance between two points. Defaults to 0.5.
        border_length (float, optional): length of the border. Defaults to 2.

    Returns:
        (np.ndarray, np.ndarray, mp.ndarray): grid points in the x, y and z axis
    """

    pmin, pmax = get_boundaries(
        atomic_positions, border_length=border_length)
    npts = np.ceil((pmax-pmin) / resolution).astype('int')

    x = np.linspace(pmin[0], pmax[0], npts[0])
    y = np.linspace(pmin[1], pmax[1], npts[1])
    z = np.linspace(pmin[2], pmax[2], npts[2])

    return (x, y, z)


def interpolator_reg_grid(func, x, y, z):
    """Computes the interpolation function

    Args:
        func (callable):  compute the value of the funtion to interpolate
        x (np.ndarray): grid points in the x direction
        y (np.ndarray): grid points in the y direction
        z (np.ndarray): grid points in the z direction

    Returns:
        callable: interpolation function
    """
    nx, ny, nz = len(x), len(y), len(z)
    grid = np.stack(np.meshgrid(
        z, y, x, indexing='ij')).T.reshape(-1, 3)[:, [2, 1, 0]]

    data = func(grid).detach().numpy()
    data = data.reshape(nx, ny, nz, -1)
    return RegularGridInterpolator((x, y, z),
                                   data,
                                   method='linear',
                                   bounds_error=False,
                                   fill_value=0.)


def interpolate_reg_grid(interpfunc, pos):
    """Interpolate the  funtion

    Args:
        interpfunc (callable): function to interpolate the data points
        pos (torch.tensor): positions of the walkers Nbatch x 3*Nelec

    Returns:
        torch.tensor: interpolated values of the function evaluated at pos
    """
    nbatch = pos.shape[0]
    nelec = pos.shape[1]//3
    ndim = 3

    data = interpfunc(pos.reshape(
        nbatch, nelec, ndim).detach().numpy())

    return torch.as_tensor(data)


def is_even(x):
    """return true if x is even."""
    return x//2*2 == x


def logspace(n, length):
    """returns a 1d array of logspace between -length and +length."""
    k = np.log(length+1)/np.log(10)
    if is_even(n):
        x = np.logspace(0.01, k, n//2)-1
        return np.concatenate((-x[::-1], x[1:]))
    else:
        x = np.logspace(0.0, k, n//2+1)-1
        return np.concatenate((-x[::-1], x[1:]))


def get_log_grid(atomic_positions, n=6, length=2., border_length=2.):
    """Computes a logarithmic grid

    Args:
        atomic_positions (list, np.ndarray, torch.tensor): positions of the atoms
        n (int, optional): number of points in each axis around each atom. Defaults to 6.
        length (float, optional): absolute value of the max distance from the atom. Defaults to 2.
        border_length (float, optional): length of the border. Defaults to 2.

    Returns:
        np.ndanrray: grid points (Npts,3)
    """

    x, y, z = np.stack(get_boundaries(
        atomic_positions, border_length=border_length)).T
    grid_pts = np.stack(np.meshgrid(
        x, y, z, indexing='ij')).T.reshape(-1, 3)

    x = logspace(n, length)
    pts = np.stack(np.meshgrid(x, x, x, indexing='ij')
                   ).T.reshape(-1, 3)

    for pos in atomic_positions:
        _tmp = pts + pos
        if grid_pts is None:
            grid_pts = _tmp
        else:
            grid_pts = np.concatenate((grid_pts, _tmp))
    return grid_pts


def interpolator_irreg_grid(func, grid_pts):
    """compute a linear ND interpolator

    Args:
        func (callable):  compute the value of the funtion to interpolate
        grid_pts (np.ndarray): grid points in the x direction

    Returns:
        callable: interpolation function
    """

    return LinearNDInterpolator(grid_pts,
                                func(grid_pts),
                                fill_value=0.)


def interpolate_irreg_grid(interpfunc, pos):
    """Interpolate the  funtion

    Args:
        interpfunc (callable): function to interpolate the data points
        pos (torch.tensor): positions of the walkers Nbatch x 3*Nelec

    Returns:
        torch.tensor: interpolated values of the function evaluated at pos
    """

    nbatch, nelec, ndim = pos.shape[0], pos.shape[1]//3, 3
    return torch.as_tensor(interpfunc(pos.reshape(nbatch, nelec, ndim)))



from copy import deepcopy
from time import time

import torch
from torch.utils.data import DataLoader
from qmctorch.utils import (DataSet, Loss,
                            OrthoReg, add_group_attr,
                            dump_to_hdf5)

from .. import log
from .solver_base import SolverBase


class Solver(SolverBase):

    def __init__(self, wf=None, sampler=None, optimizer=None,
                 scheduler=None, output=None, rank=0):
        """Basic QMC solver

        Args:
            wf (qmctorch.WaveFunction, optional): wave function. Defaults to None.
            sampler (qmctorch.sampler, optional): Sampler. Defaults to None.
            optimizer (torch.optim, optional): optimizer. Defaults to None.
            scheduler (torch.optim, optional): scheduler. Defaults to None.
            output (str, optional): hdf5 filename. Defaults to None.
            rank (int, optional): rank of he process. Defaults to 0.
        """
        SolverBase.__init__(self, wf, sampler,
                            optimizer, scheduler, output, rank)

        self.set_params_requires_grad()

        self.configure(track=['local_energy'], freeze=None,
                       loss='energy', grad='manual',
                       ortho_mo=False, clip_loss=False,
                       resampling={'mode': 'update',
                                   'resample_every': 1,
                                   'nstep_update': 25})

    def configure(self, track=None, freeze=None,
                  loss=None, grad=None,
                  ortho_mo=None, clip_loss=False,
                  resampling=None):
        """Configure the solver

        Args:
            track (list, optional): list of observable to track. Defaults to ['local_energy'].
            freeze ([type], optional): list of parameters to freeze. Defaults to None.
            loss(str, optional): merhod to compute the loss: variance or energy.
                                  Defaults to 'energy'.
            grad (str, optional): method to compute the gradients: 'auto' or 'manual'.
                                  Defaults to 'auto'.
            ortho_mo (bool, optional): apply regularization to orthogonalize the MOs.
                                       Defaults to False.
            clip_loss (bool, optional): Clip the loss values at +/- X std. X defined in Loss
                                        as clip_num_std (default 5)
                                        Defaults to False.
        """

        # set the parameters we want to optimize/freeze
        self.set_params_requires_grad()
        self.freeze_params_list = freeze
        self.freeze_parameters(freeze)

        # track the observable we want
        if track is not None:
            self.track_observable(track)

        # define the grad calulation
        if grad is not None:
            self.grad_method = grad
            self.evaluate_gradient = {
                'auto': self.evaluate_grad_auto,
                'manual': self.evaluate_grad_manual}[grad]

        # resampling of the wave function
        if resampling is not None:
            self.configure_resampling(**resampling)

        # get the loss
        if loss is not None:
            self.loss = Loss(self.wf, method=loss, clip=clip_loss)
            self.loss.use_weight = (
                self.resampling_options.resample_every > 1)

        # orthogonalization penalty for the MO coeffs
        if ortho_mo is not None:
            self.ortho_mo = ortho_mo
            self.ortho_loss = OrthoReg()

    def set_params_requires_grad(self, wf_params=True, geo_params=False):
        """Configure parameters for wf opt."""

        # opt all wf parameters
        self.wf.ao.bas_exp.requires_grad = wf_params
        self.wf.ao.bas_coeffs.requires_grad = wf_params

        for param in self.wf.mo.parameters():
            param.requires_grad = wf_params

        self.wf.fc.weight.requires_grad = wf_params

        for param in self.wf.jastrow.parameters():
            param.requires_grad = wf_params

        # no opt the atom positions
        self.wf.ao.atom_coords.requires_grad = geo_params

    def freeze_parameters(self, freeze):
        """Freeze the optimization of specified params.

        Args:
            freeze (list): list of param to freeze
        """
        if freeze is not None:
            if not isinstance(freeze, list):
                freeze = [freeze]

            for name in freeze:
                if name.lower() == 'ci':
                    self.wf.fc.weight.requires_grad = False

                elif name.lower() == 'mo':
                    for param in self.wf.mo.parameters():
                        param.requires_grad = False

                elif name.lower() == 'ao':
                    self.wf.ao.bas_exp.requires_grad = False
                    self.wf.ao.bas_coeffs.requires_grad = False

                elif name.lower() == 'jastrow':
                    for param in self.wf.jastrow.parameters():
                        param.requires_grad = False

                else:
                    opt_freeze = ['ci', 'mo', 'ao', 'jastrow']
                    raise ValueError(
                        'Valid arguments for freeze are :', opt_freeze)

    def save_sampling_parameters(self, pos):
        """ save the sampling params."""
        self.sampler._nstep_save = self.sampler.nstep
        self.sampler._ntherm_save = self.sampler.ntherm
        self.sampler._nwalker_save = self.sampler.walkers.nwalkers

        if self.resampling_options.mode == 'update':
            self.sampler.ntherm = -1
            self.sampler.nstep = self.resampling_options.nstep_update
            self.sampler.walkers.nwalkers = pos.shape[0]
            self.sampler.nwalkers = pos.shape[0]

    def restore_sampling_parameters(self):
        """restore sampling params to their original values."""
        self.sampler.nstep = self.sampler._nstep_save
        self.sampler.ntherm = self.sampler._ntherm_save
        self.sampler.walkers.nwalkers = self.sampler._nwalker_save
        self.sampler.nwalkers = self.sampler._nwalker_save

    def geo_opt(self, nepoch, geo_lr=1e-2, batchsize=None,
                nepoch_wf_init=100, nepoch_wf_update=50,
                hdf5_group='geo_opt', chkpt_every=None, tqdm=False):
        """optimize the geometry of the molecule

        Args:
            nepoch (int): Number of optimziation step
            batchsize (int, optional): Number of sample in a mini batch.
                                       If None, all samples are used.
                                       Defaults to Never.
            hdf5_group (str, optional): name of the hdf5 group where to store the data.
                                        Defaults to 'geo_opt'.
            chkpt_every (int, optional): save a checkpoint every every iteration.
                                         Defaults to half the number of epoch
        """

        # save the optimizer used for the wf params
        opt_wf = deepcopy(self.opt)
        opt_wf.lpos_needed = self.opt.lpos_needed

        # create the optmizier for the geo opt
        opt_geo = torch.optim.SGD(self.wf.parameters(), lr=geo_lr)
        opt_geo.lpos_needed = False

        # save the grad method
        eval_grad_wf = self.evaluate_gradient

        # log data
        self.prepare_optimization(batchsize, None, tqdm)
        self.log_data_opt(nepoch, 'geometry optimization')

        # init the traj
        xyz = [self.wf.geometry(None)]

        # initial wf optimization
        self.set_params_requires_grad(wf_params=True,
                                      geo_params=False)
        self.freeze_parameters(self.freeze_params_list)
        self.run_epochs(nepoch_wf_init)

        # iterations over geo optim
        for n in range(nepoch):

            # make one step geo optim
            self.set_params_requires_grad(wf_params=False,
                                          geo_params=True)
            self.opt = opt_geo
            self.evaluate_gradient = self.evaluate_grad_auto
            self.run_epochs(1)
            xyz.append(self.wf.geometry(None))

            # make a few wf optim
            self.set_params_requires_grad(wf_params=True,
                                          geo_params=False)
            self.freeze_parameters(self.freeze_params_list)
            self.opt = opt_wf
            self.evaluate_gradient = eval_grad_wf

            cumulative_loss = self.run_epochs(nepoch_wf_update)

            # save checkpoint file
            if chkpt_every is not None:
                if (n > 0) and (n % chkpt_every == 0):
                    self.save_checkpoint(n, cumulative_loss)

        # restore the sampler number of step
        self.restore_sampling_parameters()

        # dump
        self.observable.geometry = xyz
        self.save_data(hdf5_group)

        return self.observable

    def run(self, nepoch, batchsize=None,
            hdf5_group='wf_opt', chkpt_every=None, tqdm=False):
        """Run a wave function optimization

        Args:
            nepoch (int): Number of optimziation step
            batchsize (int, optional): Number of sample in a mini batch.
                                       If None, all samples are used.
                                       Defaults to Never.
            hdf5_group (str, optional): name of the hdf5 group where to store the data.
                                        Defaults to 'wf_opt'.
            chkpt_every (int, optional): save a checkpoint every every iteration.
                                         Defaults to half the number of epoch
        """

        # prepare the optimization
        self.prepare_optimization(batchsize, chkpt_every, tqdm)
        self.log_data_opt(nepoch, 'wave function optimization')

        # run the epochs
        self.run_epochs(nepoch)

        # restore the sampler number of step
        self.restore_sampling_parameters()

        # dump
        self.save_data(hdf5_group)

        return self.observable

    def prepare_optimization(self, batchsize, chkpt_every, tqdm=False):
        """Prepare the optimization process

        Args:
            batchsize (int or None): batchsize
            chkpt_every (int or none): save a chkpt file every
        """

        # sample the wave function
        pos = self.sampler(self.wf.pdf, with_tqdm=tqdm)

        # handle the batch size
        if batchsize is None:
            batchsize = len(pos)

        # change the number of steps/walker size
        self.save_sampling_parameters(pos)

        # create the data loader
        self.dataset = DataSet(pos)
        self.dataloader = DataLoader(
            self.dataset, batch_size=batchsize)

        for ibatch, data in enumerate(self.dataloader):
            self.store_observable(data, ibatch=ibatch)

        # chkpt
        self.chkpt_every = chkpt_every

    def save_data(self, hdf5_group):
        """Save the data to hdf5.

        Args:
            hdf5_group (str): name of group in the hdf5 file
        """
        self.observable.models.last = dict(self.wf.state_dict())

        hdf5_group = dump_to_hdf5(
            self.observable, self.hdf5file, hdf5_group)

        add_group_attr(self.hdf5file, hdf5_group, {'type': 'opt'})

    def run_epochs(self, nepoch):
        """Run a certain number of epochs

        Args:
            nepoch (int): number of epoch to run
        """

        # init the loss in case we have nepoch=0
        cumulative_loss = 0

        # loop over the epoch
        for n in range(nepoch):

            tstart = time()
            log.info('')
            log.info('  epoch %d' % n)

            cumulative_loss = 0

            # loop over the batches
            for ibatch, data in enumerate(self.dataloader):

                # port data to device
                lpos = data.to(self.device)

                # get the gradient
                loss, eloc = self.evaluate_gradient(lpos)
                cumulative_loss += loss

                # check for nan
                if torch.isnan(eloc).any():
                    log.info('Error : Nan detected in local energy')
                    return cumulative_loss

                # optimize the parameters
                self.optimization_step(lpos)

                # observable
                self.store_observable(
                    lpos, local_energy=eloc, ibatch=ibatch)

            # save the model if necessary
            if n == 0 or cumulative_loss < min_loss:
                min_loss = cumulative_loss
                self.observable.models.best = dict(
                    self.wf.state_dict())

            # save checkpoint file
            if self.chkpt_every is not None:
                if (n > 0) and (n % self.chkpt_every == 0):
                    self.save_checkpoint(n, cumulative_loss)

            self.print_observable(cumulative_loss, verbose=False)

            # resample the data
            self.dataset.data = self.resample(n, self.dataset.data)

            # scheduler step
            if self.scheduler is not None:
                self.scheduler.step()

            log.info('  epoch done in %1.2f sec.' % (time()-tstart))

        return cumulative_loss

    def evaluate_grad_auto(self, lpos):
        """Evaluate the gradient using automatic differentiation

        Args:
            lpos (torch.tensor): sampling points

        Returns:
            tuple: loss values and local energies
        """

        # compute the loss
        loss, eloc = self.loss(lpos)

        # add mo orthogonalization if required
        if self.wf.mo.weight.requires_grad and self.ortho_mo:
            loss += self.ortho_loss(self.wf.mo.weight)

        # compute local gradients
        self.opt.zero_grad()
        loss.backward()

        return loss, eloc

    def evaluate_grad_manual(self, lpos):
        """Evaluate the gradient using low variance expression

        Args:
            lpos ([type]): [description]

        Args:
            lpos (torch.tensor): sampling points

        Returns:
            tuple: loss values and local energies
        """

        # determine if we need the grad of eloc
        no_grad_eloc = True
        if self.wf.kinetic_method == 'auto':
            no_grad_eloc = False

        if self.wf.jastrow.requires_autograd:
            no_grad_eloc = False

        if self.loss.method in ['energy', 'weighted-energy']:

            # Get the gradient of the total energy
            # dE/dk = < (dpsi/dk)/psi (E_L - <E_L >) >
            

            # compute local energy and wf values
            _, eloc = self.loss(lpos, no_grad=no_grad_eloc)
            psi = self.wf(lpos)
            norm = 1. / len(psi)

            # evaluate the prefactor of the grads
            weight = eloc.clone()
            weight -= torch.mean(eloc)
            weight /= psi
            weight *= 2.
            weight *= norm

            # compute the gradients
            self.opt.zero_grad()
            psi.backward(weight)

            return torch.mean(eloc), eloc

        else:
            raise ValueError(
                'Manual gradient only for energy minimization')

    def log_data_opt(self, nepoch, task):
        """Log data for the optimization."""
        log.info('')
        log.info('  Optimization')
        log.info('  Task                :', task)
        log.info(
            '  Number Parameters   : {0}', self.wf.get_number_parameters())
        log.info('  Number of epoch     : {0}', nepoch)
        log.info(
            '  Batch size          : {0}', self.sampler.get_sampling_size())
        log.info('  Loss function       : {0}', self.loss.method)
        log.info('  Clip Loss           : {0}', self.loss.clip)
        log.info('  Gradients           : {0}', self.grad_method)
        log.info(
            '  Resampling mode     : {0}', self.resampling_options.mode)
        log.info(
            '  Resampling every    : {0}', self.resampling_options.resample_every)
        log.info(
            '  Resampling steps    : {0}', self.resampling_options.nstep_update)
        log.info(
            '  Output file         : {0}', self.hdf5file)
        log.info(
            '  Checkpoint every    : {0}', self.chkpt_every)
        log.info('')



import torch
from torch import nn
from .scaling import (get_scaled_distance,
                      get_der_scaled_distance,
                      get_second_der_scaled_distance)


class ElectronNucleiDistance(nn.Module):

    def __init__(self, nelec, atomic_pos, ndim=3, scale=False, scale_factor=0.6):
        """Computes the electron-nuclei distances

        .. math::
            r_{iA} = \\sqrt{ (x_i-x_A)^2 + (y_i-y_A)^2 + (z_i-z_A)^2}

        Args:
            nelec (int): number of electrons
            atomic_pos (tensor): positions of the atoms
            ndim (int): number of spatial dimensions
            scale(bool, optional): return scaled values, Defaults to False
            scale_factor(float, optional): value of the scale factor,
                                           Defaults to 0.6

        Examples::
            >>> endist = ElectronNucleiDistance(2,2)
            >>> epos = torch.tensor(500,6)
            >>> r = edist(pos)
            >>> dr = edist(pos,derivative=1)

        """

        super().__init__()
        self.nelec = nelec
        self.atoms = atomic_pos
        self.ndim = ndim
        self.scale = scale
        self.kappa = scale_factor

    def forward(self, input, derivative=0):
        """Compute the pairwise distances between electrons and atoms
        or their derivative.

        Args:
            input (torch.tesnor): position of the electron \n
                                  size : Nbatch x [Nelec x Ndim]
            derivative (int, optional): degre of the derivative. \n
                                        Defaults to 0.

        Returns:
            torch.tensor: distance (or derivative) matrix \n
                          Nbatch x Nelec x Natom if derivative = 0 \n
                          Nbatch x Ndim x  Nelec x Natom if derivative = 1,2

        """

        # get the distance matrices
        input_ = input.view(-1, self.nelec, self.ndim)
        dist = self._get_distance_quadratic(input_, self.atoms)
        dist = torch.sqrt(dist)

        if derivative == 0:
            if self.scale:
                return get_scaled_distance(self.kappa, dist)
            else:
                return dist

        elif derivative == 1:
            der_dist = self.get_der_distance(input_, dist)
            if self.scale:
                return get_der_scaled_distance(self.kappa,
                                               dist, der_dist)
            else:
                return der_dist

        elif derivative == 2:

            d2_dist = self.get_second_der_distance(input_, dist)

            if self.scale:
                der_dist = self.get_der_distance(input_, dist)
                return get_second_der_scaled_distance(self.kappa,
                                                      dist,
                                                      der_dist,
                                                      d2_dist)
            else:
                return d2_dist

    def get_der_distance(self, pos, dist):
        """Get the derivative of the electron-nuclei distance matrix

        .. math::
            \\frac{d r_{iA}}{d x_i}


        Args:
            pos (torch.tensor): positions of the electrons
                                Nbatch x Nelec x Ndim
            dist (torch.tensor): distance matrix between the elecs
                           Nbatch x Nelec x Nelec

        Returns:
            [type]: [description]
        """
        invr = (1. / dist).unsqueeze(-1)
        diff_axis = (pos.unsqueeze(-1) -
                     self.atoms.T).transpose(2, 3)
        return (diff_axis * invr).permute(0, 3, 1, 2)

    def get_second_der_distance(self, pos, dist):
        """Get the derivative of the electron-nuclei distance matrix

        .. math::
            \\frac{d^2 r_{iA}}{d x_i^2}

        Args:
            pos (torch.tensor): positions of the electrons
                                Nbatch x Nelec x Ndim
            dist (torch.tensor): distance matrix between the elecs
                           Nbatch x Nelec x Nelec

        Returns:
            [type]: [description]
        """
        invr3 = (1. / (dist**3)).unsqueeze(1)
        diff_axis = pos.transpose(1, 2).unsqueeze(3)
        diff_axis = (diff_axis - self.atoms.T.unsqueeze(1))**2

        diff_axis = diff_axis[:, [
            [1, 2], [2, 0], [0, 1]], ...].sum(2)

        return (diff_axis * invr3)

    @staticmethod
    def _get_distance_quadratic(elec_pos, atom_pos):
        """Compute the distance following a quadratic expansion

        Arguments:
            input {torch.tensor} -- electron position [nbatch x nelec x ndim]

        Returns:
            torch.tensor -- distance matrices nbatch x nelec x ndim]
        """
        norm = (elec_pos**2).sum(-1).unsqueeze(-1)
        norm_atom = (atom_pos**2).sum(-1).unsqueeze(-1).T
        dist = (norm + norm_atom - 2.0 * elec_pos@atom_pos.T)
        return dist





import numpy as np
import torch
from .slater_jastrow import SlaterJastrow

from .jastrows.elec_elec.kernels.pade_jastrow_kernel import PadeJastrowKernel as PadeJastrowKernelElecElec
from .jastrows.jastrow_factor_combined_terms import JastrowFactorCombinedTerms
from .jastrows.elec_nuclei.kernels.pade_jastrow_kernel import PadeJastrowKernel as PadeJastrowKernelElecNuc


class SlaterManyBodyJastrow(SlaterJastrow):

    def __init__(self, mol, configs='ground_state',
                 kinetic='jacobi',
                 jastrow_kernel={
                     'ee': PadeJastrowKernelElecElec,
                     'en': PadeJastrowKernelElecNuc,
                     'een': None},
                 jastrow_kernel_kwargs={
                     'ee': {},
                     'en': {},
                     'een': {}},
                 cuda=False,
                 include_all_mo=True):
        """Slater Jastrow wave function with many body Jastrow factor

        .. math::
            \\Psi(R_{at}, r) = J(r)\\sum_n c_n D^\\uparrow_n(r^\\uparrow)D^\\downarrow_n(r^\\downarrow)

        with

        .. math::
            J(r) = \\exp\\left( K_{ee}(r) + K_{en}(R_{at},r) + K_{een}(R_{at}, r) \\right)    

        with the different kernels representing electron-electron, electron-nuclei and electron-electron-nuclei terms

        Args:
            mol (Molecule): a QMCTorch molecule object
            configs (str, optional): defines the CI configurations to be used. Defaults to 'ground_state'.
                - ground_state : only the ground state determinant in the wave function
                - single(n,m) : only single excitation with n electrons and m orbitals 
                - single_double(n,m) : single and double excitation with n electrons and m orbitals
                - cas(n, m) : all possible configuration using n eletrons and m orbitals                   
            kinetic (str, optional): method to compute the kinetic energy. Defaults to 'jacobi'.
                - jacobi : use the Jacobi formula to compute the kinetic energy 
                - auto : use automatic differentiation to compute the kinetic energy
            jastrow_kernel (dict, optional) : different Jastrow kernels for the different terms. 
                By default only electron-electron and electron-nuclei terms are used
            jastrow_kernel_kwargs (dict, optional) : keyword arguments for the jastrow kernels contructor
            cuda (bool, optional): turns GPU ON/OFF  Defaults to False.
            include_all_mo (bool, optional): include either all molecular orbitals or only the ones that are
                                             popualted in the configs. Defaults to False
        Examples::
            >>> from qmctorch.scf import Molecule
            >>> from qmctorch.wavefunction import SlaterManyBodyJastrow
            >>> mol = Molecule('h2o.xyz', calculator='adf', basis = 'dzp')
            >>> wf = SlaterManyBodyJastrow(mol, configs='cas(2,2)')
        """

        super().__init__(mol, configs, kinetic, None, {}, cuda, include_all_mo)

        # process the Jastrow
        if jastrow_kernel is not None:

            for k in ['ee', 'en', 'een']:
                if k not in jastrow_kernel.keys():
                    jastrow_kernel[k] = None
                if k not in jastrow_kernel_kwargs.keys():
                    jastrow_kernel_kwargs[k] = None

            self.use_jastrow = True
            self.jastrow_type = 'JastrowFactorCombinedTerms'

            self.jastrow = JastrowFactorCombinedTerms(
                self.mol.nup, self.mol.ndown,
                torch.as_tensor(self.mol.atom_coords),
                jastrow_kernel=jastrow_kernel,
                jastrow_kernel_kwargs=jastrow_kernel_kwargs,
                cuda=cuda)

            if self.cuda:
                for term in self.jastrow.jastrow_terms:
                    term = term.to(self.device)

        self.log_data()



from copy import deepcopy

import matplotlib.pyplot as plt
import numpy as np
import torch
from scipy.optimize import curve_fit
from torch import nn

import torch
from .. import log
from ..utils import register_extra_attributes
from .orbitals.atomic_orbitals import AtomicOrbitals
from .pooling.orbital_configurations import OrbitalConfigurations
from .pooling.slater_pooling import SlaterPooling
from .wf_base import WaveFunction


class SlaterJastrowBase(WaveFunction):

    def __init__(self, mol,
                 configs='ground_state',
                 kinetic='jacobi',
                 cuda=False,
                 include_all_mo=True):
        """Implementation of the QMC Network.

        Args:
            mol (Molecule): a QMCTorch molecule object
            configs (str, optional): defines the CI configurations to be used. Defaults to 'ground_state'.
                - ground_state : only the ground state determinant in the wave function
                - single(n,m) : only single excitation with n electrons and m orbitals 
                - single_double(n,m) : single and double excitation with n electrons and m orbitals
                - cas(n, m) : all possible configuration using n eletrons and m orbitals                   
            kinetic (str, optional): method to compute the kinetic energy. Defaults to 'jacobi'.
                - jacobi : use the Jacobi formula to compute the kinetic energy 
                - auto : use automatic differentiation to compute the kinetic energy
            cuda (bool, optional): turns GPU ON/OFF  Defaults to False.
            include_all_mo (bool, optional): include either all molecular orbitals or only the ones that are
                                             popualted in the configs. Defaults to False
        """

        super(SlaterJastrowBase, self).__init__(
            mol.nelec, 3, kinetic, cuda)

        # check for cuda
        if not torch.cuda.is_available and self.cuda:
            raise ValueError('Cuda not available, use cuda=False')

        # check for conf/mo size
        if not include_all_mo and configs.startswith('cas('):
            raise ValueError(
                'CAS calculation only possible with include_all_mo=True')

        # number of atoms
        self.mol = mol
        self.atoms = mol.atoms
        self.natom = mol.natom

        # define the SD we want
        self.orb_confs = OrbitalConfigurations(mol)
        self.configs_method = configs
        self.configs = self.orb_confs.get_configs(configs)
        self.nci = len(self.configs[0])
        self.highest_occ_mo = torch.stack(self.configs).max()+1

        # define the atomic orbital layer
        self.ao = AtomicOrbitals(mol, cuda)

        # define the mo layer
        self.include_all_mo = include_all_mo
        self.nmo_opt = mol.basis.nmo if include_all_mo else self.highest_occ_mo
        self.mo_scf = nn.Linear(
            mol.basis.nao, self.nmo_opt, bias=False)
        self.mo_scf.weight = self.get_mo_coeffs()
        self.mo_scf.weight.requires_grad = False
        if self.cuda:
            self.mo_scf.to(self.device)

        # define the mo mixing layer
        # self.mo = nn.Linear(mol.basis.nmo, self.nmo_opt, bias=False)
        self.mo = nn.Linear(self.nmo_opt, self.nmo_opt, bias=False)
        self.mo.weight = nn.Parameter(
            torch.eye(self.nmo_opt, self.nmo_opt))
        if self.cuda:
            self.mo.to(self.device)

        # jastrow
        self.jastrow_type = None
        self.use_jastrow = False

        #  define the SD pooling layer
        self.pool = SlaterPooling(self.configs_method,
                                  self.configs, mol, cuda)

        # define the linear layer
        self.fc = nn.Linear(self.nci, 1, bias=False)
        self.fc.weight.data.fill_(0.)
        self.fc.weight.data[0][0] = 1.

        if self.cuda:
            self.fc = self.fc.to(self.device)

        self.kinetic_method = kinetic
        if kinetic == 'jacobi':
            self.kinetic_energy = self.kinetic_energy_jacobi

        gradients = 'auto'
        self.gradients_method = gradients
        if gradients == 'jacobi':
            self.gradients = self.gradients_jacobi

        if self.cuda:
            self.device = torch.device('cuda')
            self.to(self.device)

        # register the callable for hdf5 dump
        register_extra_attributes(self,
                                  ['ao', 'mo_scf',
                                   'mo', 'jastrow',
                                   'pool', 'fc'])

    def log_data(self):
        """Print information abut the wave function."""
        log.info('')
        log.info(' Wave Function')
        log.info('  Jastrow factor      : {0}', self.use_jastrow)
        if self.use_jastrow:
            log.info(
                '  Jastrow kernel      : {0}', self.jastrow_type)
        log.info('  Highest MO included : {0}', self.nmo_opt)
        log.info('  Configurations      : {0}', self.configs_method)
        log.info('  Number of confs     : {0}', self.nci)

        log.debug('  Configurations      : ')
        for ic in range(self.nci):
            cstr = ' ' + ' '.join([str(i)
                                   for i in self.configs[0][ic].tolist()])
            cstr += ' | ' + ' '.join([str(i)
                                      for i in self.configs[1][ic].tolist()])
            log.debug(cstr)

        log.info('  Kinetic energy      : {0}', self.kinetic_method)
        log.info(
            '  Number var  param   : {0}', self.get_number_parameters())
        log.info('  Cuda support        : {0}', self.cuda)
        if self.cuda:
            log.info(
                '  GPU                 : {0}', torch.cuda.get_device_name(0))

    def get_mo_coeffs(self):
        mo_coeff = torch.as_tensor(self.mol.basis.mos).type(
            torch.get_default_dtype())
        if not self.include_all_mo:
            mo_coeff = mo_coeff[:, :self.highest_occ_mo]
        return nn.Parameter(mo_coeff.transpose(0, 1).contiguous())

    def update_mo_coeffs(self):
        self.mol.atom_coords = self.ao.atom_coords.detach().numpy().tolist()
        self.mo.weight = self.get_mo_coeffs()

    def geometry(self, pos):
        """Returns the gemoetry of the system in xyz format

        Args:
            pos (torch.tensor): sampling points (Nbatch, 3*Nelec)

        Returns:
            list: list where each element is one line of the xyz file
        """
        d = []
        for iat in range(self.natom):

            xyz = self.ao.atom_coords[iat,
                                      :].cpu().detach().numpy().tolist()
            d.append(xyz)
        return d

    def gto2sto(self, plot=False):
        """Fits the AO GTO to AO STO.
            The SZ sto that have only one basis function per ao
        """

        assert(self.ao.radial_type.startswith('gto'))
        assert(self.ao.harmonics_type == 'cart')

        log.info('  Fit GTOs to STOs  : ')

        def sto(x, norm, alpha):
            """Fitting function."""
            return norm * np.exp(-alpha * np.abs(x))

        # shortcut for nao
        nao = self.mol.basis.nao

        # create a new mol and a new basis
        new_mol = deepcopy(self.mol)
        basis = deepcopy(self.mol.basis)

        # change basis to sto
        basis.radial_type = 'sto_pure'
        basis.nshells = self.ao.nao_per_atom.detach().cpu().numpy()

        # reset basis data
        basis.index_ctr = np.arange(nao)
        basis.bas_coeffs = np.ones(nao)
        basis.bas_exp = np.zeros(nao)
        basis.bas_norm = np.zeros(nao)
        basis.bas_kr = np.zeros(nao)
        basis.bas_kx = np.zeros(nao)
        basis.bas_ky = np.zeros(nao)
        basis.bas_kz = np.zeros(nao)

        # 2D fit space
        x = torch.linspace(-5, 5, 501)

        # compute the values of the current AOs using GTO BAS
        pos = x.reshape(-1, 1).repeat(1, self.ao.nbas).to(self.device)
        gto = self.ao.norm_cst * torch.exp(-self.ao.bas_exp*pos**2)
        gto = gto.unsqueeze(1).repeat(1, self.nelec, 1)
        ao = self.ao._contract(gto)[
            :, 0, :].detach().cpu().numpy()

        # loop over AOs
        for iorb in range(self.ao.norb):

            # fit AO with STO
            xdata = x.numpy()
            ydata = ao[:, iorb]
            popt, pcov = curve_fit(sto, xdata, ydata)

            # store new exp/norm
            basis.bas_norm[iorb] = popt[0]
            basis.bas_exp[iorb] = popt[1]

            # determine k values
            basis.bas_kx[iorb] = self.ao.harmonics.bas_kx[self.ao.index_ctr == iorb].unique(
            ).item()
            basis.bas_ky[iorb] = self.ao.harmonics.bas_ky[self.ao.index_ctr == iorb].unique(
            ).item()
            basis.bas_kz[iorb] = self.ao.harmonics.bas_kz[self.ao.index_ctr == iorb].unique(
            ).item()

            # plot if necessary
            if plot:
                plt.plot(xdata, ydata)
                plt.plot(xdata, sto(xdata, *popt))
                plt.show()

        # update basis in new mole
        new_mol.basis = basis

        # returns new orbital instance
        return self.__class__(new_mol, configs=self.configs_method,
                              kinetic=self.kinetic_method,
                              cuda=self.cuda,
                              include_all_mo=self.include_all_mo)

    def forward(self, x, ao=None):
        """computes the value of the wave function for the sampling points

        .. math::
            \\Psi(R) = \\sum_{n} c_n D^{u}_n(r^u) \\times D^{d}_n(r^d)

        Args:
            x (torch.tensor): sampling points (Nbatch, 3*Nelec)
            ao (torch.tensor, optional): values of the atomic orbitals (Nbatch, Nelec, Nao)

        Returns:
            torch.tensor: values of the wave functions at each sampling point (Nbatch, 1)

        Examples::
            >>> mol = Molecule('h2.xyz', calculator='adf', basis = 'dzp')
            >>> wf = SlaterJastrow(mol, configs='cas(2,2)')
            >>> pos = torch.rand(500,6)
            >>> vals = wf(pos)
        """

        raise NotImplementedError('Implement a forward method')

    def ao2mo(self, ao):
        """Get the values of the MO from the values of AO."""
        raise NotImplementedError('Implement a ao2mo method')

    def pos2mo(self, x, derivative=0):
        """Get the values of MOs from the positions

        Arguments:
            x {torch.tensor} -- positions of the electrons [nbatch, nelec*ndim]

        Keyword Arguments:
            derivative {int} -- order of the derivative (default: {0})

        Returns:
            torch.tensor -- MO matrix [nbatch, nelec, nmo]
        """
        raise NotImplementedError('Implement a get_mo_vals method')

    def kinetic_energy_jacobi(self, x,  **kwargs):
        """Compute the value of the kinetic enery using the Jacobi Formula.
        C. Filippi, Simple Formalism for Efficient Derivatives .

        .. math::
             \\frac{K(R)}{\\Psi(R)} = Tr(A^{-1} B_{kin})

        Args:
            x (torch.tensor): sampling points (Nbatch, 3*Nelec)

        Returns:
            torch.tensor: values of the kinetic energy at each sampling points
        """

        raise NotImplementedError(
            'Implement a kinetic_energy_jacobi method')

    def gradients_jacobi(self, x, pdf=False):
        """Compute the gradients of the wave function (or density) using the Jacobi Formula
        C. Filippi, Simple Formalism for Efficient Derivatives.

        .. math::
             \\frac{K(R)}{\Psi(R)} = Tr(A^{-1} B_{grad})

        Args:
            x (torch.tensor): sampling points (Nbatch, 3*Nelec)
            pdf (bool, optional) : if true compute the grads of the density

        Returns:
            torch.tensor: values of the gradients wrt the walker pos at each sampling points
        """

        raise NotImplementedError(
            'Implement a gradient_jacobi method')

    def get_gradient_operator(self, x, ao, grad_ao, mo):
        """Compute the gradient operator

        Args:
            x ([type]): [description]
            ao ([type]): [description]
            dao ([type]): [description]
        """

        raise NotImplementedError(
            'Implement a get_grad_operator method')

    def get_hessian_operator(self, x, ao, dao, d2ao,  mo):
        """Compute the Bkin matrix

        Args:
            x (torch.tensor): sampling points (Nbatch, 3*Nelec)
            mo (torch.tensor, optional): precomputed values of the MOs

        Returns:
            torch.tensor: matrix of the kinetic operator
        """

        raise NotImplementedError(
            'Implement a get_kinetic_operator method')



import numpy
import torch
from torch import nn
from ...jastrows.distance.electron_electron_distance import ElectronElectronDistance
from .orbital_dependent_backflow_kernel import OrbitalDependentBackFlowKernel


class OrbitalDependentBackFlowTransformation(nn.Module):

    def __init__(self, mol, backflow_kernel, backflow_kernel_kwargs={}, cuda=False):
        """Transform the electorn coordinates into backflow coordinates.
        see : Orbital-dependent backflow wave functions for real-space quantum Monte Carlo
        https://arxiv.org/abs/1910.07167

        .. math:
            \\bold{q}_i = \\bold{r}_i + \\sum_{j\neq i} \\eta(r_{ij})(\\bold{r}_i - \\bold{r}_j)
        """
        super().__init__()

        self.edist = ElectronElectronDistance(mol.nelec)
        self.nelec = mol.nelec
        self.nao = mol.basis.nao
        self.backflow_kernel = OrbitalDependentBackFlowKernel(
            backflow_kernel, backflow_kernel_kwargs, mol, cuda)
        self.ndim = 3

        self.cuda = cuda
        self.device = torch.device('cpu')
        if self.cuda:
            self.device = torch.device('cuda')

    def forward(self, pos, derivative=0):

        if derivative == 0:
            return self._backflow(pos)

        elif derivative == 1:
            return self._backflow_derivative(pos)

        elif derivative == 2:
            return self._backflow_second_derivative(pos)

        else:
            raise ValueError(
                'Derivative of the backflow transformation must be 0, 1 or 2')

    def _backflow(self, pos):
        """Computes the backflow transformation

        .. math:
            \\bold{q}_i = \\bold{r}_i + \\sum_{j\neq i} \\eta(r_{ij})(\\bold{r}_i - \\bold{r}_j)

        Args:
            pos(torch.tensor): original positions Nbatch x[Nelec*Ndim]

        Returns:
            torch.tensor: transformed positions Nbatch x[Nelec*Ndim]
        """

        nbatch = pos.shape[0]

        # compute the difference
        # Nbatch x 1 x Nelec x Nelec x 3
        delta_ee = self.edist.get_difference(
            pos.reshape(nbatch, self.nelec, self.ndim)).unsqueeze(1)

        # compute the backflow function
        # Nbatch x Nao x Nelec x Nelec x 1
        bf_kernel = self.backflow_kernel(
            self.edist(pos)).unsqueeze(-1)
        nao = bf_kernel.shape[self.backflow_kernel.stack_axis]

        # update pos
        pos = pos.reshape(nbatch, 1, self.nelec, self.ndim) + \
            (bf_kernel * delta_ee).sum(3)

        # retrurn Nbatch x Nao x Nelec*Ndim
        return pos.reshape(nbatch, nao, self.nelec*self.ndim)

    def _backflow_derivative(self, pos):
        r"""Computes the derivative of the backflow transformation
           wrt the original positions of the electrons

        .. math::
            \\bold{q}_i = \\bold{r}_i + \\sum_{j\\neq i} \\eta(r_{ij})(\\bold{r}_i - \\bold{r}_j)

        .. math::
            \\frac{d q_i}{d x_k} = \\delta_{ik}(1 + \\sum_{j\\neq i} \\frac{d \\eta(r_ij)}{d x_i}(x_i-x_j) + \\eta(r_ij)) +
                                   \\delta_{i\\neq k}(-\\frac{d \\eta(r_ik)}{d x_k}(x_i-x_k) - \\eta(r_ik))

        Args:
            pos(torch.tensor): orginal positions of the electrons Nbatch x[Nelec*Ndim]

        Returns:
            torch.tensor: d q_{i}/d x_k with:
                          q_{i} bf position of elec i
                          x_k original coordinate of the kth elec
                          Nelec x  Nbatch x Nelec x Norb x Ndim
        """

        # ee dist matrix : Nbatch x  Nelec x Nelec
        ree = self.edist(pos)
        nbatch, nelec, _ = ree.shape

        # derivative ee dist matrix : Nbatch x 1 x 3 x Nelec x Nelec
        # dr_ij / dx_i = - dr_ij / dx_j
        dree = self.edist(pos, derivative=1).unsqueeze(1)

        # difference between elec pos
        # Nbatch, 1, 3, Nelec, Nelec
        delta_ee = self.edist.get_difference(
            pos.reshape(nbatch, nelec, 3)).permute(0, 3, 1, 2).unsqueeze(1)

        # backflow kernel : Nbatch x Nao x Nelec x Nelec
        bf = self.backflow_kernel(ree)
        nao = bf.shape[self.backflow_kernel.stack_axis]

        # (d eta(r_ij) / d r_ij) (d r_ij/d beta_i)
        # derivative of the back flow kernel : Nbatch x Nao x 3 x Nelec x Nelec
        dbf = self.backflow_kernel(ree, derivative=1).unsqueeze(2)
        dbf = dbf * dree

        # (d eta(r_ij) / d beta_i) (alpha_i - alpha_j)
        # Nbatch x Nao x 3 x 3 x Nelec x Nelec
        dbf_delta_ee = dbf.unsqueeze(2) * delta_ee.unsqueeze(3)

        # compute the delta_ij * (1 + sum k \neq i eta(rik))
        # Nbatch x Nao x Nelec x Nelec (diagonal matrix)
        delta_ij_bf = torch.diag_embed(
            1 + bf.sum(-1), dim1=-1, dim2=-2)

        # eye 3x3 in 1x3x3x1x1
        eye_mat = torch.eye(3, 3).view(
            1, 1, 3, 3, 1, 1).to(self.device)

        # compute the delta_ab * delta_ij * (1 + sum k \neq i eta(rik))
        # Nbatch x Ndim x Ndim x Nelec x Nelec (diagonal matrix)
        delta_ab_delta_ij_bf = eye_mat * \
            delta_ij_bf.view(nbatch, nao, 1, 1, nelec, nelec)

        # compute sum_k df(r_ik)/dbeta_i (alpha_i - alpha_k)
        # Nbatch x Nao x Ndim x Ndim x Nelec x Nelec
        delta_ij_sum = torch.diag_embed(
            dbf_delta_ee.sum(-1), dim1=-1, dim2=-2)

        # compute delta_ab * f(rij)
        delta_ab_bf = eye_mat * \
            bf.view(nbatch, nao, 1, 1, nelec, nelec)

        # return Nbatch x Ndim(alpha) x Ndim(beta) x Nelec(i) x Nelec(j)
        # nbatch d alpha_i / d beta_j
        return delta_ab_delta_ij_bf + delta_ij_sum - dbf_delta_ee - delta_ab_bf

    def _backflow_second_derivative(self, pos):
        r"""Computes the second derivative of the backflow transformation
           wrt the original positions of the electrons

        .. math::
            \\bold{q}_i = \\bold{r}_i + \\sum_{j\\neq i} \\eta(r_{ij})(\\bold{r}_i - \\bold{r}_j)

        .. math::
            \\frac{d q_i}{d x_k} = \\delta_{ik}(1 + \\sum_{j\\neqi} \\frac{d \\eta(r_ij)}{d x_i} + \\eta(r_ij)) +
                                   \\delta_{i\\neq k}(-\\frac{d \\eta(r_ik)}{d x_k} - \\eta(r_ik))

        .. math::
            \\frac{d ^ 2 q_i}{d x_k ^ 2} = \\delta_{ik}(\\sum_{j\\neqi} \\frac{d ^ 2 \\eta(r_ij)}{d x_i ^ 2} + 2 \\frac{d \\eta(r_ij)}{d x_i}) +
                                       - \\delta_{i\\neq k}(\\frac{d ^ 2 \\eta(r_ik)}{d x_k ^ 2} + \\frac{d \\eta(r_ik)}{d x_k})

        Args:
            pos(torch.tensor): orginal positions of the electrons Nbatch x[Nelec*Ndim]

        Returns:
            torch.tensor: d q_{i}/d x_k with:
                          q_{i} bf position of elec i
                          x_k original coordinate of the kth elec
                          Nelec x  Nbatch x Nelec x Norb x Ndim
        """

        # ee dist matrix :
        # Nbatch x  Nelec x Nelec
        ree = self.edist(pos)
        nbatch, nelec, _ = ree.shape

        # difference between elec pos
        # Nbatch, 1, 3, Nelec, Nelec
        delta_ee = self.edist.get_difference(
            pos.reshape(nbatch, nelec, 3)).permute(0, 3, 1, 2).unsqueeze(1)

        # derivative ee dist matrix  d r_{ij} / d x_i
        # Nbatch x 1 x 3 x Nelec x Nelec
        dree = self.edist(pos, derivative=1).unsqueeze(1)

        # derivative ee dist matrix :  d2 r_{ij} / d2 x_i
        # Nbatch x 1 x 3 x Nelec x Nelec
        d2ree = self.edist(pos, derivative=2).unsqueeze(1)

        # derivative of the back flow kernel : d eta(r_ij)/d r_ij
        # Nbatch x Nao x 1 x Nelec x Nelec
        dbf = self.backflow_kernel(ree, derivative=1).unsqueeze(2)
        nao = dbf.shape[self.backflow_kernel.stack_axis]

        # second derivative of the back flow kernel : d2 eta(r_ij)/d2 r_ij
        # Nbatch x Nao x 1 x Nelec x Nelec
        d2bf = self.backflow_kernel(ree, derivative=2).unsqueeze(2)

        # (d^2 eta(r_ij) / d r_ij^2) (d r_ij/d x_i)^2
        # + (d eta(r_ij) / d r_ij) (d^2 r_ij/d x_i^2)
        # Nbatch x Nao x 3 x Nelec x Nelec
        d2bf = (d2bf * dree * dree) + (dbf * d2ree)

        # (d eta(r_ij) / d r_ij) (d r_ij/d x_i)
        # Nbatch x Nao x 3 x Nelec x Nelec
        dbf = dbf * dree

        # eye matrix in dim x dim
        eye_mat = torch.eye(3, 3).reshape(
            1, 1, 3, 3, 1, 1).to(self.device)

        # compute delta_ij delta_ab 2 sum_k dbf(ik) / dbeta_i
        term1 = 2 * eye_mat * \
            torch.diag_embed(
                dbf.sum(-1), dim1=-1, dim2=-2).reshape(nbatch, nao, 1, 3, nelec, nelec)

        # (d2 eta(r_ij) / d2 beta_i) (alpha_i - alpha_j)
        # Nbatch x Nao x 3 x 3 x Nelec x Nelec
        d2bf_delta_ee = d2bf.unsqueeze(2) * delta_ee.unsqueeze(3)

        # compute sum_k d2f(r_ik)/d2beta_i (alpha_i - alpha_k)
        # Nbatch x Nao x Ndim x Ndim x Nelec x Nelec
        term2 = torch.diag_embed(
            d2bf_delta_ee.sum(-1), dim1=-1, dim2=-2)

        # compute delta_ab * df(rij)/dbeta_j
        term3 = 2 * eye_mat * \
            dbf.reshape(nbatch, nao, 1, 3, nelec, nelec)

        # return Nbatch x Ndim(alpha) x Ndim(beta) x Nelec(i) x Nelec(j)
        # nbatch d2 alpha_i / d2 beta_j
        return term1 + term2 + d2bf_delta_ee + term3
